import os
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import torch
from torch import nn
from torch.nn import CrossEntropyLoss
from torch.utils.data import Dataset
from transformers import (
    PreTrainedModel,
    PreTrainedTokenizerBase,
    Trainer,
    TrainerCallback,
    TrainingArguments,
)
from transformers.trainer_utils import PREFIX_CHECKPOINT_DIR

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.constants import DISABLE_LOSS_LABEL
from turbo_alignment.trainers.multigpu import MultiGPUCherryPicksTrainer

logger = get_project_logger()


class TrainerCustomSave(MultiGPUCherryPicksTrainer):
    def _save_checkpoint(self, model, trial, metrics=None):
        logger.info('Running custom _save_checkpoint')
        checkpoint_folder = f'{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}'
        run_dir = self._get_output_dir(trial=trial)
        output_dir = Path(os.path.join(run_dir, checkpoint_folder))

        if metrics is not None and self.args.metric_for_best_model is not None:
            metric_to_check = self.args.metric_for_best_model
            if not metric_to_check.startswith('eval_'):
                metric_to_check = f'eval_{metric_to_check}'
            metric_value = metrics[metric_to_check]
            operator = np.greater if self.args.greater_is_better else np.less
            if (
                self.state.best_metric is None
                or self.state.best_model_checkpoint is None
                or operator(metric_value, self.state.best_metric)
            ):
                self.state.best_metric = metric_value
                self.state.best_model_checkpoint = output_dir

        if self.args.should_save:
            self._rotate_checkpoints(use_mtime=True, output_dir=run_dir)

        (output_dir / 'projections').mkdir(parents=True, exist_ok=True)
        (output_dir / 'adapter').mkdir(parents=True, exist_ok=True)
        (output_dir / 'tokenizer').mkdir(parents=True, exist_ok=True)

        if isinstance(model, torch.nn.DataParallel):
            torch.save(
                model.module.modality_adapters.state_dict(), output_dir / 'projections' / 'modality_adapters.pt'
            )
            model.module.language_model.save_pretrained(output_dir / 'adapter')
        else:
            torch.save(model.modality_adapters.state_dict(), output_dir / 'projections' / 'modality_adapters.pt')
            model.language_model.save_pretrained(output_dir / 'adapter')

        self.tokenizer.save_pretrained(output_dir / 'tokenizer')


@dataclass
class MultimodalTrainingArguments(TrainingArguments):
    average_log_prob: bool = False
    gamma: float = 0.0001


class MultimodalTrainer(Trainer):
    def __init__(
        self,
        model: Union[PreTrainedModel | nn.Module],
        data_collator: Callable,
        args: MultimodalTrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        model_init: Optional[Callable[[], PreTrainedModel]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        optimizers: Tuple[Optional[torch.optim.Optimizer], Optional[torch.optim.lr_scheduler.LambdaLR]] = (
            None,
            None,
        ),
        preprocess_logits_for_metrics: Optional[Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
        **kwargs,
    ):
        self.data_collator = data_collator

        self._stored_metrics: Dict[str, Dict[str, List[float]]] = defaultdict(lambda: defaultdict(list))
        self._average_log_prob = args.average_log_prob
        self._gamma = args.gamma

        super().__init__(
            model,
            args,
            data_collator,
            train_dataset,
            eval_dataset,
            tokenizer,
            model_init,
            None,
            callbacks,
            optimizers,
            preprocess_logits_for_metrics,
            **kwargs,
        )

    def concatenated_inputs(self, batch: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """
        Объединяем в один прогон батч с нормальной attention маской и с той, где в modality tokens стоят нолики
        """

        no_modality_attention_mask = torch.logical_and(
            batch['attention_mask'],
            torch.logical_not(batch['modality_tokens_mask']),
        )

        concatenated_batch: Dict[str, torch.Tensor] = {
            'input_ids': batch['input_ids'].repeat(2, 1).to(self.accelerator.device),
            'modality_tokens_mask': batch['modality_tokens_mask'].repeat(2, 1).to(self.accelerator.device),
            'attention_mask': torch.cat(
                [
                    batch['attention_mask'],
                    no_modality_attention_mask,
                ],
                dim=0,
            ).to(self.accelerator.device),
            'labels': batch['input_ids'].repeat(2, 1).to(self.accelerator.device),
            'modality_inputs': batch['modality_inputs'] + batch['modality_inputs'],
        }

        return concatenated_batch

    def multimodal_loss(
        self,
        logits: torch.Tensor,
        logps: torch.Tensor,
        ignore_modality_logps: torch.Tensor,
        labels: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        seq_len = (labels != DISABLE_LOSS_LABEL).sum(1)
        # modality_indifference = seq_len / (torch.abs(logps.sum() - ignore_modality_logps.sum()) + self._gamma)
        modality_indifference = seq_len / (logps.sum() - ignore_modality_logps.sum() + self._gamma)

        shift_logits = logits[..., :-1, :].contiguous()
        shift_labels = labels[..., 1:].contiguous()

        ce_loss_fct = CrossEntropyLoss(reduction='none')
        loss = ce_loss_fct(shift_logits.view(-1, shift_logits.size(-1)), shift_labels.view(-1))

        sample_mean_ce_loss = loss.view(logits.size(0), -1).mean(dim=1)

        loss = (modality_indifference * sample_mean_ce_loss).mean()
        return loss, modality_indifference

    @staticmethod
    def _get_batch_logps(
        logits: torch.Tensor,
        labels: torch.Tensor,
        average_log_prob: bool = False,
    ) -> torch.Tensor:
        if logits.shape[:-1] != labels.shape:
            raise ValueError('Logits (batch and sequence length dim) and labels must have the same shape.')

        labels = labels[:, 1:].clone()
        logits = logits[:, :-1, :]
        loss_mask = labels != DISABLE_LOSS_LABEL

        labels[labels == DISABLE_LOSS_LABEL] = 0

        per_token_logps = torch.gather(logits.log_softmax(-1), dim=2, index=labels.unsqueeze(2)).squeeze(2)

        if average_log_prob:
            return (per_token_logps * loss_mask).sum(-1) / loss_mask.sum(-1)
        return (per_token_logps * loss_mask).sum(-1)

    def concatenated_forward(
        self, model: nn.Module, batch: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        concatenated_batch = self.concatenated_inputs(batch)
        all_logits = model(**{k: v for k, v in concatenated_batch.items() if k != 'labels'}).logits.to(torch.float32)

        all_logps = self._get_batch_logps(
            all_logits,
            concatenated_batch['labels'],
            average_log_prob=self._average_log_prob,
        )
        chosen_idxs = batch['input_ids'].shape[0]

        logps = all_logps[:chosen_idxs]
        no_modality_logps = all_logps[chosen_idxs:]

        logits = all_logits[:chosen_idxs]
        no_modality_logits = all_logits[chosen_idxs:]
        return logps, no_modality_logps, logits, no_modality_logits

    def get_batch_metrics(
        self,
        batch: Dict[str, Any],
        train_eval: Literal['train', 'eval'] = 'train',
    ) -> Tuple[torch.Tensor, Dict[str, float]]:
        def _compute_cpu_mean(tensor: torch.Tensor) -> float:
            return tensor.detach().cpu().mean().item()

        assert 'modality_tokens_mask' in batch
        assert 'attention_mask' in batch
        assert 'input_ids' in batch
        assert 'labels' in batch

        metrics = {}

        (
            logps,
            ignore_modality_logps,
            logits,
            _,
        ) = self.concatenated_forward(self.model, batch)

        loss, modality_indifference = self.multimodal_loss(
            logits,
            logps,
            ignore_modality_logps,
            batch['labels'],
        )

        prefix = 'eval_' if train_eval == 'eval' else ''
        metrics[f'{prefix}logits/modality_indifference'] = _compute_cpu_mean(modality_indifference)
        metrics[f'{prefix}logits/logps'] = _compute_cpu_mean(logps)
        metrics[f'{prefix}logits/ignore_modality_logps'] = _compute_cpu_mean(ignore_modality_logps)

        return loss, metrics

    def compute_loss(
        self,
        _model: Union[PreTrainedModel, nn.Module],
        inputs: Dict[str, Union[torch.Tensor, Any]],
        return_outputs: bool = False,
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, Dict[str, float]]]:
        loss, metrics = self.get_batch_metrics(inputs, train_eval='train')

        if self.accelerator.is_main_process:
            self.store_metrics(metrics, train_eval='train')
        if return_outputs:
            return loss, metrics
        return loss

    def prediction_step(
        self,
        model: Union[PreTrainedModel, nn.Module],
        inputs: Dict[str, Union[torch.Tensor, Any]],
        prediction_loss_only: bool,
        ignore_keys: Optional[List[str]] = None,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor], Optional[torch.Tensor]]:
        if ignore_keys is None:
            if hasattr(model, 'config'):
                ignore_keys = getattr(model.config, 'keys_to_ignore_at_inference', [])
            else:
                ignore_keys = []

        with torch.inference_mode():
            loss, metrics = self.get_batch_metrics(inputs, train_eval='eval')

        if self.accelerator.is_main_process:
            self.store_metrics(metrics, train_eval='eval')
        if prediction_loss_only:
            return loss.detach(), None, None

        logits_dict = {
            'logits_test/chosen': metrics['logits_test/chosen'],
            'logits_test/rejected': metrics['logits_test/rejected'],
        }
        logits = tuple(v for k, v in logits_dict.items() if k not in ignore_keys)
        logits = torch.stack(logits).mean(axis=1)  # type: ignore[call-overload, arg-type]
        labels = torch.zeros(logits.shape[0])

        return loss.detach(), logits, labels

    def store_metrics(self, metrics: Dict[str, float], train_eval: Literal['train', 'eval'] = 'train') -> None:
        for key, value in metrics.items():
            self._stored_metrics[train_eval][key].append(value)

    def log(self, logs: Dict[str, float]) -> None:
        train_eval = 'train' if 'loss' in logs else 'eval'
        for key, metrics in self._stored_metrics[train_eval].items():
            logs[key] = torch.tensor(metrics).mean().item()
        del self._stored_metrics[train_eval]
        return super().log(logs)  # pylint: disable=no-member

    def _save_checkpoint(self, model, trial, metrics=None):
        logger.info('Running custom _save_checkpoint')
        checkpoint_folder = f'{PREFIX_CHECKPOINT_DIR}-{self.state.global_step}'
        run_dir = self._get_output_dir(trial=trial)
        output_dir = Path(os.path.join(run_dir, checkpoint_folder))

        if metrics is not None and self.args.metric_for_best_model is not None:
            metric_to_check = self.args.metric_for_best_model
            if not metric_to_check.startswith('eval_'):
                metric_to_check = f'eval_{metric_to_check}'
            metric_value = metrics[metric_to_check]
            operator = np.greater if self.args.greater_is_better else np.less
            if (
                self.state.best_metric is None
                or self.state.best_model_checkpoint is None
                or operator(metric_value, self.state.best_metric)
            ):
                self.state.best_metric = metric_value
                self.state.best_model_checkpoint = output_dir

        (output_dir / 'projections').mkdir(parents=True, exist_ok=True)
        (output_dir / 'language_model').mkdir(parents=True, exist_ok=True)
        (output_dir / 'tokenizer').mkdir(parents=True, exist_ok=True)

        torch.save(model.image_linear_projection.state_dict(), output_dir / 'projections' / 'image_connector.pt')
        torch.save(model.audio_linear_projection.state_dict(), output_dir / 'projections' / 'audio_connector.pt')

        model.language_model.save_pretrained(output_dir / 'language_model')
        self.tokenizer.save_pretrained(output_dir / 'tokenizer')



from copy import deepcopy

import numpy as np

from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.dataset.sampling.rm import SamplingRMDataset
from turbo_alignment.pipelines.sampling.rm import BaseSamplingStrategyWithRM
from turbo_alignment.settings.pipelines.sampling import RSOSamplingSettings


class RSOSamplingStrategy(BaseSamplingStrategyWithRM[RSOSamplingSettings]):
    @staticmethod
    def __sample_for_single_record(
        experiment_settings: RSOSamplingSettings,
        record: SamplingDatasetRecord,
        candidates: dict[str, float],
    ) -> SamplingDatasetRecord:
        accepted: list[str] = []

        while len(accepted) < experiment_settings.N and candidates:
            max_reward = max(candidates.values())

            for c, r in candidates.items():
                u = np.random.uniform()
                if u >= np.exp((r - max_reward) / experiment_settings.beta):
                    continue

                accepted.append(c)

                if len(accepted) == experiment_settings.N:
                    break

            to_remove = []
            for c in candidates:
                if c not in accepted:
                    to_remove.append(c)
            for c in to_remove:
                candidates.pop(c)

        sampled_record = deepcopy(record)
        sampled_record.rewards = candidates

        answers = []
        for answer in sampled_record.answers:
            if answer.id in candidates:
                answers.append(answer)
        sampled_record.answers = answers
        return sampled_record

    @staticmethod
    def _sample(
        experiment_settings: RSOSamplingSettings,
        dataset: SamplingRMDataset,
        rewards: dict[str, dict[str, float]],
    ) -> list[SamplingDatasetRecord]:
        sampled_records: list[SamplingDatasetRecord] = []

        for record_id, record in dataset.original_records_map.items():
            sampled_records.append(
                RSOSamplingStrategy.__sample_for_single_record(
                    experiment_settings=experiment_settings,
                    record=record,
                    candidates=rewards[record_id],
                )
            )

        return sampled_records



from pathlib import Path
from typing import Sequence

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.chat import ChatMultiDatasetSettings
from turbo_alignment.settings.pipelines.inference.chat import ChatGenerationSettings
from turbo_alignment.settings.rag_model import RAGPreTrainedModelSettings
from turbo_alignment.settings.tf.tokenizer import TokenizerSettings


class RAGSingleModelInferenceSettings(ExtraFieldsNotAllowedBaseModel):
    model_settings: RAGPreTrainedModelSettings
    tokenizer_settings: TokenizerSettings
    batch: int = 1

    generation_settings: list[ChatGenerationSettings]


class RAGInferenceExperimentSettings(ExtraFieldsNotAllowedBaseModel):
    inference_settings: Sequence[RAGSingleModelInferenceSettings]

    dataset_settings: ChatMultiDatasetSettings
    save_path: Path



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class TokenizerSettings(ExtraFieldsNotAllowedBaseModel):
    use_fast: bool = False
    trust_remote_code: bool = False

    tokenizer_path: str | None = None

    tokenizer_kwargs: dict = {}



import os
import re
from pathlib import Path

from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)
from transformers.trainer_utils import get_last_checkpoint

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.s3.checkpoints_handler import S3CheckpointsHandler
from turbo_alignment.settings.s3 import CheckpointUploaderCallbackParameters

logger = get_project_logger()


class CheckpointUploaderCallback(TrainerCallback):
    def __init__(self, parameters: CheckpointUploaderCallbackParameters, s3_handler: S3CheckpointsHandler) -> None:
        self._parameters = parameters
        self._s3_handler = s3_handler

    def on_train_begin(
        self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs
    ) -> TrainerControl:
        logger.info(f"💾 Saving experiment metadata to {Path(args.output_dir) / 'experiment_metadata.json'}")

        last_checkpoint_directory = get_last_checkpoint(args.output_dir)
        last_checkpoint_basename = str(last_checkpoint_directory).rsplit('/', maxsplit=1)[-1]

        self._s3_handler.upload_local_file(
            target_directory=Path(self._parameters.checkpoints_directory) / last_checkpoint_basename,
            file_path=Path(args.output_dir) / 'experiment_metadata.json',
        )

        return control

    def on_save(
        self, args: TrainingArguments, state: TrainerState, control: TrainerControl, **kwargs
    ) -> TrainerControl:
        last_checkpoint_directory = get_last_checkpoint(args.output_dir)
        last_checkpoint_basename = str(last_checkpoint_directory).rsplit('/', maxsplit=1)[-1]

        if last_checkpoint_directory is None:
            return control

        self._upload_checkpoint(checkpoint_directory=Path(last_checkpoint_directory))

        self._s3_handler.upload_local_file(
            target_directory=Path(self._parameters.checkpoints_directory) / last_checkpoint_basename,
            file_path=Path(args.output_dir) / 'experiment.config',
        )

        local_checkpoints_names = self._retrieve_recent_checkpoint_names(output_directory=Path(args.output_dir))

        remote_checkpoints_names = self._s3_handler.list_folders_keys(folder=self._parameters.checkpoints_directory)

        self._remove_old_checkpoints(
            local_checkpoints_names=local_checkpoints_names, remote_checkpoints_names=remote_checkpoints_names
        )

        return control

    def _upload_checkpoint(self, checkpoint_directory: Path) -> None:
        last_checkpoint_basename = str(checkpoint_directory).rsplit('/', maxsplit=1)[-1]
        s3_key = f'{self._parameters.checkpoints_directory}/{last_checkpoint_basename}'
        logger.info(f'🥭 Start to upload checkpoint {checkpoint_directory} to s3 with key {s3_key}')
        self._s3_handler.upload_local_files(
            path=checkpoint_directory,
            target_directory=s3_key,
            save_optimizer_states=self._parameters.save_optimizer_states,
        )
        logger.info(f'🥝 Successfully uploaded checkpoint {checkpoint_directory} to s3 with key {s3_key}')

    @staticmethod
    def _retrieve_recent_checkpoint_names(output_directory: Path) -> list[str]:
        ordering_and_checkpoint_path = []
        glob_checkpoints = [str(x) for x in output_directory.glob('checkpoint-*') if x.is_dir()]
        for path in glob_checkpoints:
            regex_match = re.match('.*(checkpoint-([0-9]+))', path)
            if regex_match is not None and regex_match.groups() is not None:
                ordering_and_checkpoint_path.append((int(regex_match.groups()[1]), regex_match.groups()[0]))

        checkpoints_sorted = sorted(ordering_and_checkpoint_path)
        checkpoints_sorted = [checkpoint[1] for checkpoint in checkpoints_sorted]

        logger.info(f'🍓 Most recent checkpoints are {checkpoints_sorted}')
        return checkpoints_sorted

    def _remove_old_checkpoints(self, local_checkpoints_names: list[str], remote_checkpoints_names: list[str]) -> None:
        for remote_checkpoint_path in remote_checkpoints_names:
            if remote_checkpoint_path not in local_checkpoints_names:
                logger.info(
                    f'''
                            🥭 Start to remove {remote_checkpoint_path} since it is
                            not in {local_checkpoints_names}'''
                )
                self._s3_handler.remove_directory(
                    os.path.join(self._parameters.checkpoints_directory, remote_checkpoint_path)
                )
                logger.info('🥝 Successfully removed %s from s3', remote_checkpoint_path)



from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.dataset.chat.models import ChatMessage


class ClassificationDatasetRecord(DatasetRecord):
    messages: list[ChatMessage]
    label: int | None



from typing import Callable

import torch
from torch import nn
from torch.utils.data import Dataset
from transformers import (
    DataCollator,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    TrainerCallback,
    TrainingArguments,
)

from turbo_alignment.trainers.multigpu import MultiGPUCherryPicksTrainer


class CustomLossTrainer(MultiGPUCherryPicksTrainer):
    def __init__(
        self,
        model: PreTrainedModel | nn.Module,
        args: TrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        custom_loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        data_collator: DataCollator,
        tokenizer: PreTrainedTokenizerBase | None = None,
        callbacks: list[TrainerCallback] | None = None,
        model_init: Callable[[], PreTrainedModel] | None = None,
        **kwargs,
    ):
        self.custom_loss = custom_loss
        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            model_init=model_init,
            callbacks=callbacks,
            **kwargs,
        )

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        Modified original version, without manual label smoothing
        """
        if 'labels' in inputs:
            labels = inputs.pop('labels')
        else:
            raise ValueError('No labels provided in the inputs')

        outputs = model(**inputs)
        logits = outputs['logits'] if isinstance(outputs, dict) else outputs[0]

        loss = self.custom_loss(logits, labels)

        return (loss, outputs) if return_outputs else loss



from collections import defaultdict
from pathlib import Path
from typing import Any, overload

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.dataset.base import AlignmentDataset
from turbo_alignment.dataset.chat import TrainChatDataset
from turbo_alignment.dataset.chat.models import (
    ChatDatasetRecord,
    ChatMessage,
    ChatMessageRole,
)
from turbo_alignment.dataset.registry import SamplingRMDatasetTypeRegistry
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.datasets.base import DatasetSourceSettings
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings


@SamplingRMDatasetTypeRegistry.register(DatasetStrategy.INFERENCE)
class SamplingRMDataset(AlignmentDataset[SamplingDatasetRecord]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: ChatDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
        read: bool = True,
    ) -> None:
        settings.keep_end = True
        self._chat_dataset = TrainChatDataset(
            settings=settings,
            source=source,
            tokenizer=tokenizer,
            read=False,
        )
        super().__init__(source=source, settings=settings, tokenizer=tokenizer)
        self.settings: ChatDatasetSettings = settings

        if read:
            self._read()

    def convert_records(self, records: list[SamplingDatasetRecord]) -> list[dict[str, Any] | None]:
        chat_records = []

        for record in records:
            prompt_messages = [ChatMessage(role=message.role, content=message.content) for message in record.messages]

            answer_messages = [
                ChatMessage(role=ChatMessageRole.BOT, content=answer.content.removesuffix('</RS>'))
                for answer in record.answers
            ]

            for msg in answer_messages:
                chat_records.append(ChatDatasetRecord(id=record.id, messages=prompt_messages + [msg]))

        tokenized_chats = self._chat_dataset.convert_records(chat_records)

        tokenized_records: dict[str, Any] = defaultdict(dict)
        total_answers_counter = 0
        for record in records:
            tokenized_records[record.id]['id'] = record.id
            tokenized_records[record.id]['rewards'] = record.rewards
            tokenized_records[record.id]['answers'] = {}

            for answer in record.answers:
                if tokenized_chats[total_answers_counter] is not None:
                    tokenized_records[record.id]['answers'][answer.id] = tokenized_chats[total_answers_counter]
                total_answers_counter += 1

        return [tokenized_records[record.id] for record in records if tokenized_records[record.id] is not None]

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[SamplingDatasetRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[SamplingDatasetRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[SamplingDatasetRecord]:
        if isinstance(records, Path):
            return [SamplingDatasetRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [SamplingDatasetRecord(**record) for record in records]
        raise NotImplementedError



import torch

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import KLSettings
from turbo_alignment.metrics.utils import logprobs_from_logits
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.KL)
class KLMetric(Metric):
    def __init__(self, settings: KLSettings) -> None:
        super().__init__(settings=settings)
        self._settings: KLSettings = settings

    def compute(self, **kwargs) -> list[MetricResults]:
        answer_tokens_ids: list[torch.Tensor] = kwargs.get('answer_tokens_ids', None)
        logits: list[torch.Tensor] = kwargs.get('logits', None)
        metrics_kwargs = kwargs.get('metrics_kwargs', {})
        ref_logits: list[torch.Tensor] = metrics_kwargs.get(self._settings.ref_logits_type, None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if answer_tokens_ids is None:
            raise ValueError('answer_tokens_ids should not be None')
        if logits is None:
            raise ValueError('logits should not be None')
        if ref_logits is None:
            raise ValueError(f'{self._settings.ref_logits_type} should not be None')

        ref_logprobs = self._logprobs(ref_logits, answer_tokens_ids)
        logprobs = self._logprobs(logits, answer_tokens_ids)

        element_wise_diversity_scores = [
            ElementWiseScores(
                label=dataset_name + '@@' + f'kl_with_{self._settings.ref_logits_type}',
                values=[(logprob - ref_logprob).mean().item() for logprob, ref_logprob in zip(logprobs, ref_logprobs)],
            )
        ]

        return [
            MetricResults(element_wise_scores=element_wise_diversity_scores, need_average=need_average)
            for need_average in self._settings.need_average
        ]

    @staticmethod
    def _logprobs(logits: list[torch.Tensor], labels: list[torch.Tensor]) -> list[torch.Tensor]:
        logprobs = []
        for item_logits, item_labels in zip(logits, labels):
            logprobs.append(logprobs_from_logits(logits=item_logits[:, :-1, :].cpu(), labels=item_labels[:, 1:].cpu()))
        return logprobs



from turbo_alignment.common.registry.from_params import FromParams
from turbo_alignment.common.registry.lazy import Lazy
from turbo_alignment.common.registry.params import Params
from turbo_alignment.common.registry.registrable import Registrable



import torch
from peft import PeftModel

from turbo_alignment.common.tf.loaders import load_model, load_tokenizer
from turbo_alignment.settings.pipelines.common.merge_adapters_to_base import (
    MergeAdaptersToBaseModelSettings,
)


def peft_to_base_model(settings: MergeAdaptersToBaseModelSettings) -> None:
    tokenizer = load_tokenizer(settings.tokenizer_settings, settings.model_settings)
    base_model = load_model(settings.model_settings, tokenizer)

    first_weight = base_model.model.layers[0].self_attn.q_proj.weight
    first_weight_old = first_weight.clone()

    lora_model = PeftModel.from_pretrained(
        base_model,
        settings.adapter_path,
        device_map='auto',
        torch_dtype=base_model.dtype,
    )

    assert torch.allclose(first_weight_old, first_weight)

    lora_model = lora_model.merge_and_unload()
    lora_model.train(False)

    assert not torch.allclose(first_weight_old, first_weight)

    lora_model.save_pretrained(
        settings.save_path,
        max_shard_size=settings.max_shard_size,
    )
    tokenizer.save_pretrained(settings.save_path)



from accelerate import Accelerator
from transformers import PreTrainedModel, PreTrainedTokenizerBase

from turbo_alignment.cherry_picks.chat import ChatCherryPickCallback
from turbo_alignment.dataset.chat import InferenceChatDataset
from turbo_alignment.generators.rag import RagGenerator
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults


class RagCherryPickCallback(ChatCherryPickCallback):
    def _get_dataset_metrics(
        self,
        dataset: InferenceChatDataset,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> list[MetricResults]:
        accelerator: Accelerator = kwargs.get('accelerator', None)

        generator = RagGenerator(
            model=model,
            tokenizer=tokenizer,
            transformers_settings=self._generator_transformers_settings,
            custom_generation_settings=self._custom_generation_settings,
            accelerator=accelerator,
        )

        generations = generator.generate_from_dataset(dataset)

        prompts = [dataset[i]['prompt'] for i in range(len(dataset))]
        answers, docs, scores = [], [], []
        for gen in generations:
            ans_group_enum = [f'{ind+1}: {ans.content}' for ind, ans in enumerate(gen.answers)]
            doc_group_enum = [f'{ind+1}: {doc}' for ind, doc in enumerate(gen.documents)]
            scores_group_enum = [f'{ind+1}: {round(s, 2)}' for ind, s in enumerate(gen.doc_scores)]
            answers.append('\n\n'.join(ans_group_enum))
            docs.append('\n\n'.join(doc_group_enum))
            scores.append('\n\n'.join(scores_group_enum))

        metric_outputs = [
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'prompt', values=prompts)]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'answers', values=answers)]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'documents', values=docs)]
            ),
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset.source.name + '@@' + 'scores', values=scores)]
            ),
        ]

        for metric in self._metrics:
            metric_results = metric.compute(dataset=dataset, generations=generations)
            metric_outputs.extend(metric_results)

        return metric_outputs



# mypy: ignore-errors
# pylint: skip-file
import copy
import json
import logging
import os
import zlib
from collections import OrderedDict
from collections.abc import MutableMapping
from itertools import chain
from os import PathLike
from typing import Any, Dict, Iterable, List, Optional, Set, TypeVar, Union

try:
    from _jsonnet import evaluate_file, evaluate_snippet
except ImportError:

    def evaluate_file(filename: str, **_kwargs) -> str:
        logger.warning(f'error loading _jsonnet (this is expected on Windows), treating {filename} as plain json')
        with open(filename, 'r', 'utf-8') as evaluation_file:
            return evaluation_file.read()

    def evaluate_snippet(_filename: str, expr: str, **_kwargs) -> str:
        logger.warning('error loading _jsonnet (this is expected on Windows), treating snippet as plain json')
        return expr


from turbo_alignment.common.registry.checks import ConfigurationError
from turbo_alignment.common.registry.file_utils import cached_path

logger = logging.getLogger(__name__)


def infer_and_cast(value: Any):
    if isinstance(value, (int, float, bool)):
        return value
    if isinstance(value, list):
        return [infer_and_cast(item) for item in value]
    if isinstance(value, dict):
        return {key: infer_and_cast(item) for key, item in value.items()}
    if isinstance(value, str):
        if value.lower() == 'true':
            return True
        if value.lower() == 'false':
            return False
        try:
            return int(value)
        except ValueError:
            pass
        try:
            return float(value)
        except ValueError:
            return value
    raise ValueError(f'cannot infer type of {value}')


def _is_encodable(value: str) -> bool:
    return (value == '') or (value.encode('utf-8', 'ignore') != b'')


def _environment_variables() -> Dict[str, str]:
    return {key: value for key, value in os.environ.items() if _is_encodable(value)}


T = TypeVar('T', dict, list)


def with_overrides(original: T, overrides_dict: Dict[str, Any], prefix: str = '') -> T:
    merged: T
    keys: Union[Iterable[str], Iterable[int]]
    if isinstance(original, list):
        merged = [None] * len(original)
        keys = range(len(original))
    elif isinstance(original, dict):
        merged = {}
        keys = chain(original.keys(), (k for k in overrides_dict if '.' not in k and k not in original))
    else:
        if prefix:
            raise ValueError(
                f"overrides for '{prefix[:-1]}.*' expected list or dict in original, "
                f'found {type(original)} instead'
            )
        raise ValueError(f'expected list or dict, found {type(original)} instead')

    used_override_keys: Set[str] = set()
    for key in keys:
        if str(key) in overrides_dict:
            merged[key] = copy.deepcopy(overrides_dict[str(key)])
            used_override_keys.add(str(key))
        else:
            overrides_subdict = {}
            for o_key in overrides_dict:
                if o_key.startswith(f'{key}.'):
                    overrides_subdict[o_key[len(f'{key}.') :]] = overrides_dict[o_key]
                    used_override_keys.add(o_key)
            if overrides_subdict:
                merged[key] = with_overrides(original[key], overrides_subdict, prefix=prefix + f'{key}.')
            else:
                merged[key] = copy.deepcopy(original[key])

    unused_override_keys = [prefix + key for key in set(overrides_dict.keys()) - used_override_keys]
    if unused_override_keys:
        raise ValueError(f'overrides dict contains unused keys: {unused_override_keys}')

    return merged


def parse_overrides(serialized_overrides: str, ext_vars: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    if serialized_overrides:
        ext_vars = {**_environment_variables(), **(ext_vars or {})}

        return json.loads(evaluate_snippet('', serialized_overrides, ext_vars=ext_vars))
    return {}


def _is_dict_free(obj: Any) -> bool:
    if isinstance(obj, dict):
        return False
    if isinstance(obj, list):
        return all(_is_dict_free(item) for item in obj)
    return True


class Params(MutableMapping):
    DEFAULT = object()

    def __init__(self, params: Dict[str, Any], history: str = '') -> None:
        self.params = _replace_none(params)
        self.history = history

    def pop(self, key: str, default: Any = DEFAULT, keep_as_dict: bool = False) -> Any:
        if default is self.DEFAULT:
            try:
                value = self.params.pop(key)
            except KeyError:
                msg = f'key "{key}" is required'
                if self.history:
                    msg += f' at location "{self.history}"'
                raise ConfigurationError(msg)
        else:
            value = self.params.pop(key, default)

        if keep_as_dict or _is_dict_free(value):
            logger.info(f'{self.history}{key} = {value}')
            return value
        return self._check_is_dict(key, value)

    def pop_int(self, key: str, default: Any = DEFAULT) -> Optional[int]:
        value = self.pop(key, default)
        if value is None:
            return None
        return int(value)

    def pop_float(self, key: str, default: Any = DEFAULT) -> Optional[float]:
        value = self.pop(key, default)
        if value is None:
            return None
        return float(value)

    def pop_bool(self, key: str, default: Any = DEFAULT) -> Optional[bool]:
        value = self.pop(key, default)
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        if value == 'true':
            return True
        if value == 'false':
            return False
        raise ValueError('Cannot convert variable to bool: ' + value)

    def get(self, key: str, default: Any = DEFAULT):
        default = None if default is self.DEFAULT else default
        value = self.params.get(key, default)
        return self._check_is_dict(key, value)

    def pop_choice(
        self,
        key: str,
        choices: List[Any],
        default_to_first_choice: bool = False,
        allow_class_names: bool = True,
    ) -> Any:
        default = choices[0] if default_to_first_choice else self.DEFAULT
        value = self.pop(key, default)
        ok_because_class_name = allow_class_names and '.' in value
        if value not in choices and not ok_because_class_name:
            key_str = self.history + key
            message = (
                f'{value} not in acceptable choices for {key_str}: {choices}. '
                'You should either use the --include-package flag to make sure the correct module '
                'is loaded, or use a fully qualified class name in your config file like '
                """{"model": "my_module.models.MyModel"} to have it imported automatically."""
            )
            raise ConfigurationError(message)
        return value

    def as_dict(self, quiet: bool = False, infer_type_and_cast: bool = False):
        if infer_type_and_cast:
            params_as_dict = infer_and_cast(self.params)
        else:
            params_as_dict = self.params

        if quiet:
            return params_as_dict

        def log_recursively(parameters, history):
            for key, value in parameters.items():
                if isinstance(value, dict):
                    new_local_history = history + key + '.'
                    log_recursively(value, new_local_history)
                else:
                    logger.info(f'{history}{key} = {value}')

        log_recursively(self.params, self.history)
        return params_as_dict

    def as_flat_dict(self) -> Dict[str, Any]:
        """
        Returns the parameters of a flat dictionary from keys to values.
        Nested structure is collapsed with periods.
        """
        flat_params = {}

        def recurse(parameters, path):
            for key, value in parameters.items():
                newpath = path + [key]
                if isinstance(value, dict):
                    recurse(value, newpath)
                else:
                    flat_params['.'.join(newpath)] = value

        recurse(self.params, [])
        return flat_params

    def duplicate(self) -> 'Params':
        return copy.deepcopy(self)

    def assert_empty(self, class_name: str):
        if self.params:
            raise ConfigurationError('Extra parameters passed to {}: {}'.format(class_name, self.params))

    def __getitem__(self, key):
        if key in self.params:
            return self._check_is_dict(key, self.params[key])
        raise KeyError(str(key))

    def __setitem__(self, key, value):
        self.params[key] = value

    def __delitem__(self, key):
        del self.params[key]

    def __iter__(self):
        return iter(self.params)

    def __len__(self):
        return len(self.params)

    def _check_is_dict(self, new_history, value):
        if isinstance(value, dict):
            new_history = self.history + new_history + '.'
            return Params(value, history=new_history)
        if isinstance(value, list):
            value = [self._check_is_dict(f'{new_history}.{i}', v) for i, v in enumerate(value)]
        return value

    @classmethod
    def from_file(
        cls,
        params_file: Union[str, PathLike],
        params_overrides: Union[str, Dict[str, Any]] = '',
        ext_vars: dict = None,
    ) -> 'Params':
        if ext_vars is None:
            ext_vars = {}

        # redirect to cache, if necessary
        params_file = cached_path(params_file)
        ext_vars = {**_environment_variables(), **ext_vars}

        file_dict = json.loads(evaluate_file(params_file, ext_vars=ext_vars))

        if isinstance(params_overrides, dict):
            params_overrides = json.dumps(params_overrides)
        overrides_dict = parse_overrides(params_overrides, ext_vars=ext_vars)

        if overrides_dict:
            param_dict = with_overrides(file_dict, overrides_dict)
        else:
            param_dict = file_dict

        return cls(param_dict)

    def to_file(self, params_file: str, preference_orders: List[List[str]] = None) -> None:
        with open(params_file, 'w', 'utf-8') as handle:
            json.dump(self.as_ordered_dict(preference_orders), handle, indent=4)

    def as_ordered_dict(self, preference_orders: List[List[str]] = None) -> OrderedDict:
        params_dict = self.as_dict(quiet=True)
        if not preference_orders:
            preference_orders = []
            preference_orders.append(
                [
                    'dataset_reader',
                    'iterator',
                    'model',
                    'train_data_path',
                    'validation_data_path',
                    'test_data_path',
                    'trainer',
                    'vocabulary',
                ]
            )
            preference_orders.append(['type'])

        def order_func(key):
            order_tuple = [order.index(key) if key in order else len(order) for order in preference_orders]
            return order_tuple + [key]

        def order_dict(dictionary, order_func):
            result = OrderedDict()
            for key, val in sorted(dictionary.items(), key=lambda item: order_func(item[0])):
                result[key] = order_dict(val, order_func) if isinstance(val, dict) else val
            return result

        return order_dict(params_dict, order_func)

    def get_hash(self) -> str:
        dumped = json.dumps(self.params, sort_keys=True)
        hashed = zlib.adler32(dumped.encode())
        return str(hashed)

    def __str__(self) -> str:
        return f'{self.history}Params({self.params})'


def pop_choice(
    params: Dict[str, Any],
    key: str,
    choices: List[Any],
    default_to_first_choice: bool = False,
    history: str = '?.',
    allow_class_names: bool = True,
) -> Any:
    value = Params(params, history).pop_choice(
        key, choices, default_to_first_choice, allow_class_names=allow_class_names
    )
    return value


def _replace_none(params: Any) -> Any:
    if params == 'None':
        return None
    if isinstance(params, dict):
        for key, value in params.items():
            params[key] = _replace_none(value)
        return params
    if isinstance(params, list):
        return [_replace_none(value) for value in params]
    return params


def remove_keys_from_params(params: Params, keys: list[str] | None):
    if keys is None:
        keys = ['pretrained_file', 'initializer']
    if isinstance(params, Params):
        param_keys = params.keys()
        for key in keys:
            if key in param_keys:
                del params[key]
        for value in params.values():
            if isinstance(value, Params):
                remove_keys_from_params(value, keys)



import json
from pathlib import Path
from typing import Any


def read_json(path: Path):
    with path.open(encoding='utf8') as f:
        return json.load(f)


def write_json(data: Any, path: Path) -> None:
    with path.open('w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=4, default=str)


def read_jsonl(path: Path):
    with path.open(encoding='utf-8') as f:
        return [json.loads(line) for line in f]


def write_jsonl(records: Any, path: Path) -> None:
    with path.open('w', encoding='utf-8') as f:
        for r in records:
            f.write(json.dumps(r, ensure_ascii=False) + '\n')


def read_txt(path: Path) -> str:
    with path.open(encoding='utf-8') as f:
        return f.read()


def write_string_to_file(string: str, path: Path) -> None:
    with path.open('w', encoding='utf-8') as f:
        f.write(string)


def write_labels(labels: list[str], path: Path) -> None:
    label_str = '\n'.join(labels) + '\n'
    write_string_to_file(label_str, path)


def create_placeholder_file(path: Path) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    write_string_to_file('Placeholder', path)


def read_text_file(path: Path) -> str:
    with path.open(encoding='utf8') as f:
        return f.read()



import torch
import torch.nn.functional as F


def average_pool(last_hidden_states, attention_mask):
    last_hidden = last_hidden_states.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]


def get_question_embeddings(encoder_output, attention_mask):
    # not sure if it works best with dpr, as it uses pooler_output
    last_hidden_state = encoder_output.hidden_states[-1]
    embeddings = average_pool(last_hidden_state, attention_mask)
    return F.normalize(embeddings, p=2, dim=1)


def pad_joined(t: torch.Tensor, max_len: int, pad_token: int = 0) -> torch.Tensor:
    return torch.cat([torch.full((1, max_len - t.shape[1]), pad_token).to(t), t], dim=1)


def pad_and_stack(tensor_list: list[torch.Tensor], max_len: int, pad_token: int) -> torch.Tensor:
    padded_tensor_list = [pad_joined(t, max_len, pad_token).squeeze(0) for t in tensor_list]
    return torch.stack(padded_tensor_list)


def join_query_and_docs(
    input_ids: torch.Tensor,
    doc_input_ids: torch.Tensor,
    pad_token_id: int,
    attention_mask: torch.Tensor | None = None,
    doc_attention_mask: torch.Tensor | None = None,
    labels: torch.Tensor | None = None,
):
    # input_ids=[1, seq_length]
    # doc_input_ids=[1, doc_length]
    inputs_list, attentions_list, labels_list = [], [], []
    max_len = 0
    for i, doc in enumerate(doc_input_ids):
        mask = torch.where(doc != pad_token_id)
        doc_inputs = doc[mask].unsqueeze(0)
        joined_inputs = torch.cat([doc_inputs, input_ids[i : i + 1]], dim=1)
        inputs_list.append(joined_inputs)
        if max_len < joined_inputs.shape[1]:
            max_len = joined_inputs.shape[1]
        if attention_mask is not None and doc_attention_mask is not None:
            doc_attn_mask = doc_attention_mask[i][mask].unsqueeze(0)
            joined_attn_mask = torch.cat([doc_attn_mask, attention_mask[i : i + 1]], dim=1)
            attentions_list.append(joined_attn_mask)
        if labels is not None:
            doc_labels = torch.full(doc_inputs.shape, -100).to(labels)
            joined_labs = torch.cat([doc_labels, labels[i : i + 1]], dim=1)
            labels_list.append(joined_labs)
    joined_input_ids = pad_and_stack(inputs_list, max_len, pad_token_id)
    joined_labels, joined_attention_mask = None, None
    if attention_mask is not None:
        joined_attention_mask = pad_and_stack(attentions_list, max_len, 0)
    if labels is not None:
        joined_labels = pad_and_stack(labels_list, max_len, -100)
    return joined_input_ids, joined_attention_mask, joined_labels



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.datasets.base import MultiDatasetSettings
from turbo_alignment.settings.datasets.chat import ChatMultiDatasetSettings
from turbo_alignment.settings.datasets.classification import (
    ClassificationMultiDatasetSettings,
)
from turbo_alignment.settings.datasets.multimodal import MultimodalMultiDatasetSettings
from turbo_alignment.settings.datasets.pair_preference import (
    PairPreferenceMultiDatasetSettings,
)
from turbo_alignment.settings.generators.chat import CustomChatGenerationSettings
from turbo_alignment.settings.metric import MetricSettings
from turbo_alignment.settings.tf.generation import GeneratorTransformersSettings


class CherryPickSettings(ExtraFieldsNotAllowedBaseModel):
    dataset_settings: MultiDatasetSettings
    metric_settings: list[MetricSettings]


class RMCherryPickSettings(CherryPickSettings):
    dataset_settings: PairPreferenceMultiDatasetSettings


class ClassificationCherryPickSettings(CherryPickSettings):
    dataset_settings: ClassificationMultiDatasetSettings


class GenerationSettings(CherryPickSettings):
    generator_transformers_settings: GeneratorTransformersSettings
    custom_generation_settings: CustomChatGenerationSettings


class ChatCherryPickSettings(GenerationSettings):
    dataset_settings: ChatMultiDatasetSettings


class MultimodalCherryPickSettings(GenerationSettings):
    dataset_settings: MultimodalMultiDatasetSettings



from pathlib import Path
from typing import Any, overload

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import read_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base import BaseDataset
from turbo_alignment.dataset.chat import (
    ChatDatasetRecord,
    ChatMessage,
    TrainChatDataset,
)
from turbo_alignment.dataset.pair_preferences import (
    PairPreferenceDataset,
    PairPreferenceRecord,
)
from turbo_alignment.dataset.registry import DDPODatasetTypeRegistry
from turbo_alignment.settings.datasets.base import (
    DatasetSourceSettings,
    DatasetStrategy,
)
from turbo_alignment.settings.datasets.ddpo import DDPOMultiDatasetSettings

logger = get_project_logger()


@DDPODatasetTypeRegistry.register(DatasetStrategy.TRAIN)
class DDPODataset(BaseDataset[PairPreferenceRecord]):
    def __init__(
        self,
        source: DatasetSourceSettings,
        settings: DDPOMultiDatasetSettings,
        chat_tokenizer: PreTrainedTokenizerBase,
        rm_tokenizer: PreTrainedTokenizerBase,
        read: bool = True,
    ) -> None:
        settings.chat_settings.keep_end = True
        self._chat_dataset = TrainChatDataset(
            settings=settings.chat_settings,
            source=source,
            tokenizer=chat_tokenizer,
            read=False,
        )
        self._rm_dataset = PairPreferenceDataset(
            settings=settings.pair_preferences,
            source=source,
            tokenizer=rm_tokenizer,
        )

        super().__init__(
            settings=settings,
            source=source,
        )
        if read:
            self._read()

    def convert_records(self, records: list[PairPreferenceRecord]) -> list[dict[str, Any] | None]:
        chosen_chat_records: list[ChatDatasetRecord] = []
        rejected_chat_records: list[ChatDatasetRecord] = []
        rm_records: list[PairPreferenceRecord] = []

        for record in records:
            context = [
                ChatMessage(role=message.role, content=message.content, disable_loss=True)
                for message in record.context
            ]
            chosen = ChatMessage(role=record.answer_w.role, content=record.answer_w.content)
            rejected = ChatMessage(role=record.answer_l.role, content=record.answer_l.content)

            chosen_chat_records.append(ChatDatasetRecord(id=record.id, messages=context + [chosen]))
            rejected_chat_records.append(ChatDatasetRecord(id=record.id, messages=context + [rejected]))

            rm_records.append(PairPreferenceRecord(id=record.id, context=context, answer_w=chosen, answer_l=rejected))

        tokenized_chosen_records = self._chat_dataset.convert_records(chosen_chat_records)
        tokenized_rejected_records = self._chat_dataset.convert_records(rejected_chat_records)
        tokenized_rm_records = self._rm_dataset.convert_records(rm_records)

        output: list[dict[str, Any] | None] = []
        for record, chosen_record, rejected_record, rm_record in zip(
            records, tokenized_chosen_records, tokenized_rejected_records, tokenized_rm_records
        ):
            if not (chosen_record and rejected_record and rm_record):
                continue

            chosen_tokens = {k: v.squeeze(0) for k, v in chosen_record.items() if k != 'id'}
            rejected_tokens = {k: v.squeeze(0) for k, v in rejected_record.items() if k != 'id'}

            output.append(
                {
                    'sft_inputs_w': chosen_tokens,
                    'sft_inputs_l': rejected_tokens,
                    'rm_inputs_w': rm_record['inputs_w'],
                    'rm_inputs_l': rm_record['inputs_l'],
                }
            )

        return output

    @staticmethod
    @overload
    def _read_records(records: Path) -> list[PairPreferenceRecord]:
        ...

    @staticmethod
    @overload
    def _read_records(records: list[dict]) -> list[PairPreferenceRecord]:
        ...

    @staticmethod
    def _read_records(records) -> list[PairPreferenceRecord]:
        if isinstance(records, Path):
            return [PairPreferenceRecord(**record) for record in read_jsonl(records)]
        if isinstance(records, list):
            return [PairPreferenceRecord(**record) for record in records]
        raise NotImplementedError


def load_ddpo_datasets(
    multi_dataset_settings: DDPOMultiDatasetSettings,
    chat_tokenizer: PreTrainedTokenizerBase,
    rm_tokenizer: PreTrainedTokenizerBase,
) -> list[DDPODataset]:
    logger.info(
        f'Loading dataset {multi_dataset_settings.dataset_type} with settings:\n{multi_dataset_settings.dict()}'
    )

    datasets: list[DDPODataset] = []
    for source in multi_dataset_settings.sources:
        # Determining what type of dataset is in the nested registry by field 'strategy'

        dataset = DDPODataset(
            chat_tokenizer=chat_tokenizer,
            rm_tokenizer=rm_tokenizer,
            source=source,
            settings=multi_dataset_settings,
        )

        datasets.append(dataset)

    return datasets



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Callable, Literal

import torch
import torch.nn.functional as F
from torch import nn
from torch.utils.data import Dataset
from transformers import (
    DefaultFlowCallback,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    PrinterCallback,
    ProgressCallback,
    TrainerCallback,
    TrainerControl,
    TrainingArguments,
)
from transformers.integrations import get_reporting_integration_callbacks

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.callbacks.common import MetricsCallbackHandler
from turbo_alignment.settings.pipelines.train.dpo import SyncRefModelSettings
from turbo_alignment.trainers.dpo import DPOTrainer
from turbo_alignment.trainers.utils import prepare_model

logger = get_project_logger()


@dataclass
class KTOTrainingArguments(TrainingArguments):
    beta: float = 0.1
    sync_ref_settings: SyncRefModelSettings = field(
        default_factory=SyncRefModelSettings()
    )  # type: ignore[call-overload]
    use_ref_model: bool = True
    average_log_prob: bool = False
    undesirable_weight: float = 1.0
    desirable_weight: float = 1.33


class KTOTrainer(DPOTrainer):
    """
    Inspired by https://github.com/huggingface/trl/blob/main/trl/trainer/kto_trainer.py

    """

    def __init__(
        self,
        model: PreTrainedModel | nn.Module,
        data_collator: Callable,
        args: KTOTrainingArguments,
        train_dataset: Dataset,
        eval_dataset: Dataset,
        ref_model: PreTrainedModel | nn.Module | None = None,
        tokenizer: PreTrainedTokenizerBase | None = None,
        callbacks: list[TrainerCallback] | None = None,
        **kwargs,
    ):
        self.data_collator = data_collator

        self.average_log_prob = args.average_log_prob
        self.desirable_weight = args.desirable_weight
        self.undesirable_weight = args.undesirable_weight
        self.beta = args.beta

        self._stored_metrics: dict[str, dict[str, list[float]]] = defaultdict(lambda: defaultdict(list))

        super().__init__(
            model=model,
            args=args,
            data_collator=data_collator,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            tokenizer=tokenizer,
            callbacks=callbacks,
            **kwargs,
        )

        self.ref_model = ref_model

        if self.ref_model is not None:
            self.ref_model = prepare_model(self.ref_model, self.accelerator, self.is_deepspeed_enabled)

        default_callbacks = [DefaultFlowCallback] + get_reporting_integration_callbacks(self.args.report_to)
        callbacks = default_callbacks if callbacks is None else default_callbacks + callbacks
        self.callback_handler = MetricsCallbackHandler(
            callbacks,
            model,
            tokenizer,
            None,
            None,
            ref_model=self.ref_model,
            accelerator=self.accelerator,
        )
        self.add_callback(PrinterCallback if self.args.disable_tqdm else ProgressCallback)
        self.control: TrainerControl = self.callback_handler.on_init_end(self.args, self.state, self.control)

    def kto_loss(
        self,
        policy_chosen_logps: torch.Tensor,
        policy_rejected_logps: torch.Tensor,
        policy_KL_logps: torch.Tensor,
        reference_chosen_logps: torch.Tensor,
        reference_rejected_logps: torch.Tensor,
        reference_KL_logps: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        KL = (policy_KL_logps - reference_KL_logps).mean().detach()
        KL = self.accelerator.gather(KL).mean().clamp(min=0)

        if policy_chosen_logps.shape[0] != 0 or reference_chosen_logps.shape[0] != 0:
            chosen_logratios = policy_chosen_logps - reference_chosen_logps
            chosen_losses = 1 - F.sigmoid(self.beta * (chosen_logratios - KL))
            chosen_rewards = self.beta * chosen_logratios.detach()
        else:
            chosen_losses = torch.Tensor([]).to(self.accelerator.device)
            chosen_rewards = torch.Tensor([]).to(self.accelerator.device)

        if policy_rejected_logps.shape[0] != 0 or reference_rejected_logps.shape[0] != 0:
            rejected_logratios = policy_rejected_logps - reference_rejected_logps
            rejected_losses = 1 - F.sigmoid(self.beta * (KL - rejected_logratios))
            rejected_rewards = self.beta * rejected_logratios.detach()
        else:
            rejected_losses = torch.Tensor([]).to(self.accelerator.device)
            rejected_rewards = torch.Tensor([]).to(self.accelerator.device)

        losses = torch.cat(
            (self.desirable_weight * chosen_losses, self.undesirable_weight * rejected_losses),
            0,
        )

        return losses, chosen_rewards, rejected_rewards, KL

    def _forward(
        self, model: nn.Module, batch: dict[str, Any]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        with torch.no_grad():
            KL_logits = model(
                batch['KL_input_ids'],
                attention_mask=batch['KL_attention_mask'],
            ).logits

        completion_logits = model(
            batch['input_ids'],
            attention_mask=batch['attention_mask'],
        ).logits

        completion_logps = self._get_batch_logps(
            logits=completion_logits,
            labels=batch['labels'],
            average_log_prob=False,
        )

        KL_logps = self._get_batch_logps(
            logits=KL_logits,
            labels=batch['KL_labels'],
            average_log_prob=False,
        )

        return completion_logps, KL_logps, completion_logits, KL_logits

    def concatenated_forward(
        self, model: nn.Module, batch: dict[str, Any]
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        completion_logps, KL_logps, completion_logits, _ = self._forward(model, batch)

        if completion_logps.shape[0] != len(batch['is_desirable']):
            raise ValueError(
                'There is a mismatch between the number of examples in this batch and the number of '
                'examples for which an output sequence was predicted.'
            )

        chosen_idx = [i for i in range(completion_logps.shape[0]) if batch['is_desirable'][i]]
        rejected_idx = [i for i in range(completion_logps.shape[0]) if not batch['is_desirable'][i]]

        chosen_logps = completion_logps[chosen_idx, ...]
        rejected_logps = completion_logps[rejected_idx, ...]

        chosen_logits = completion_logits[chosen_idx, ...]
        rejected_logits = completion_logits[rejected_idx, ...]

        return (chosen_logps, rejected_logps, chosen_logits, rejected_logits, KL_logps)

    def _get_ref_logps(self, batch: dict[str, Any]) -> tuple[torch.Tensor, torch.Tensor]:
        with torch.no_grad():
            if self.ref_model is not None:
                reference_completion_logps, reference_KL_logps, *_ = self._forward(self.ref_model, batch)
            else:
                with self.accelerator.unwrap_model(self.model).disable_adapter():
                    (reference_completion_logps, reference_KL_logps, *_) = self._forward(self.model, batch)

        return reference_completion_logps, reference_KL_logps

    def get_batch_metrics(
        self,
        model: nn.Module,
        batch: dict[str, Any],
        train_eval: Literal['train', 'eval'] = 'train',
    ) -> tuple[torch.Tensor, dict[str, Any]]:
        metrics = {}
        batch = {k: (v.to(self.accelerator.device) if isinstance(v, torch.Tensor) else v) for k, v in batch.items()}

        (
            policy_chosen_logps,
            policy_rejected_logps,
            _,
            _,
            policy_KL_logps,
        ) = self.concatenated_forward(model, batch)

        with torch.no_grad():
            if self.ref_model is not None:
                reference_chosen_logps, reference_rejected_logps, _, _, reference_KL_logps = self.concatenated_forward(
                    self.ref_model, batch
                )
            else:
                with self.accelerator.unwrap_model(self.model).disable_adapter():
                    (
                        reference_chosen_logps,
                        reference_rejected_logps,
                        _,
                        _,
                        reference_KL_logps,
                    ) = self.concatenated_forward(self.model, batch)

        losses, chosen_rewards, rejected_rewards, kl = self.kto_loss(
            policy_chosen_logps=policy_chosen_logps,
            policy_rejected_logps=policy_rejected_logps,
            policy_KL_logps=policy_KL_logps,
            reference_chosen_logps=reference_chosen_logps,
            reference_rejected_logps=reference_rejected_logps,
            reference_KL_logps=reference_KL_logps,
        )

        num_chosen = torch.Tensor([len(chosen_rewards)]).to(self.accelerator.device)
        num_rejected = torch.Tensor([len(rejected_rewards)]).to(self.accelerator.device)

        all_num_chosen = self.accelerator.gather(num_chosen)
        all_num_rejected = self.accelerator.gather(num_rejected)

        prefix = 'eval_' if train_eval == 'eval' else ''

        if all_num_chosen.sum().item() > 0:
            metrics[f'{prefix}rewards/chosen'] = (
                (self.accelerator.gather(chosen_rewards.mean()) * all_num_chosen).nansum() / all_num_chosen.sum()
            ).item()
            metrics[f'{prefix}logps/chosen'] = (
                (self.accelerator.gather(policy_chosen_logps.mean()) * all_num_chosen).nansum() / all_num_chosen.sum()
            ).item()

        if all_num_rejected.sum().item() > 0:
            metrics[f'{prefix}rewards/rejected'] = (
                (self.accelerator.gather(rejected_rewards.mean()) * all_num_rejected).nansum() / all_num_rejected.sum()
            ).item()
            metrics[f'{prefix}logps/rejected'] = (
                (self.accelerator.gather(policy_rejected_logps.mean()) * all_num_rejected).nansum()
                / all_num_rejected.sum()
            ).item()

        metrics[f'{prefix}kl'] = kl.item()
        if all_num_chosen.sum().item() > 0 and all_num_rejected.sum().item() > 0:
            metrics[f'{prefix}rewards/margins'] = (
                metrics[f'{prefix}rewards/chosen'] - metrics[f'{prefix}rewards/rejected']
            )

        del (
            policy_chosen_logps,
            policy_rejected_logps,
            policy_KL_logps,
            reference_chosen_logps,
            reference_rejected_logps,
            reference_KL_logps,
            kl,
        )
        del (
            num_chosen,
            num_rejected,
            all_num_chosen,
            all_num_rejected,
            chosen_rewards,
            rejected_rewards,
            _,
        )

        return losses.nanmean(), metrics

    def store_metrics(self, metrics: dict[str, float], train_eval: Literal['train', 'eval'] = 'train') -> None:
        for key, value in metrics.items():
            self._stored_metrics[train_eval][key].append(value)
            if isinstance(value, list):
                self._stored_metrics[train_eval][key].extend(value)
            else:
                self._stored_metrics[train_eval][key].append(value)



from typing import Any

from clearml import Task

from turbo_alignment.settings.logging.clearml import ClearMLSettings


def create_clearml_task(parameters: ClearMLSettings, config: dict[str, Any] | None = None) -> Task:
    clearml_task = Task.init(
        task_name=parameters.task_name, project_name=parameters.project_name, continue_last_task=True  # FIXME?
    )

    clearml_task.connect_configuration(config, name='HyperParameters')

    return clearml_task



from collections import defaultdict
from typing import Any, Generic, Type, TypeVar

from pydantic import BaseModel


class Singleton(type):
    _instances: dict[Type, Any] = {}

    def __call__(cls, *args, **kwargs) -> Any:
        if cls not in cls._instances[cls]:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


_SettingsType = TypeVar('_SettingsType', bound=BaseModel)


class ParametrizedSingleton(type, Generic[_SettingsType]):
    _instances: dict[Type, dict[str, Any]] = defaultdict(dict)

    def __call__(cls, settings: _SettingsType) -> Any:
        dumped_settings = settings.json()
        if dumped_settings not in cls._instances[cls]:
            cls._instances[cls][dumped_settings] = super(ParametrizedSingleton, cls).__call__(settings)
        return cls._instances[cls][dumped_settings]



from typing import Optional

import torch

from turbo_alignment.modeling.imagebind import ImageBindSingleton
from turbo_alignment.modeling.imagebind.models import ImageBindSettings
from turbo_alignment.modeling.imagebind.models import (
    ModalityType as ImageBindModalityType,
)
from turbo_alignment.modeling.multimodal.encoders.audio.base import BaseAudioEncoder
from turbo_alignment.modeling.multimodal.encoders.registry import (
    ModalityEncoderRegistry,
)
from turbo_alignment.settings.modality import ModalityEncoderType


@ModalityEncoderRegistry.register(ModalityEncoderType.IMAGEBIND_AUDIO)
class ImageBindAudioModeling(BaseAudioEncoder):
    def __init__(self, imagebind_settings: Optional[ImageBindSettings] = None):
        super().__init__()
        if imagebind_settings is None:
            imagebind_settings = ImageBindSettings()
        self.model_imagebind = ImageBindSingleton(imagebind_settings).get()

    def encode(self, inputs: torch.Tensor) -> torch.Tensor:
        input_values = {ImageBindModalityType.AUDIO: inputs}
        audio_embeddings = self.model_imagebind(input_values)[ImageBindModalityType.AUDIO]
        audio_embeddings = torch.nn.functional.normalize(audio_embeddings, dim=-1)
        return audio_embeddings

    @property
    def emb_dim(self) -> int:
        return self.model_imagebind.modality_heads[ImageBindModalityType.AUDIO][-1].out_features



from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class CustomChatGenerationSettings(ExtraFieldsNotAllowedBaseModel):
    skip_special_tokens: bool = True
    remove_prompt: bool = True
    batch: int = 1



from enum import Enum
from pathlib import Path

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.tf.model import ModelTransformersSettings
from turbo_alignment.settings.tf.peft import PEFT_TYPE


class ModelType(str, Enum):
    CAUSAL = 'causal'
    SEQ2SEQ = 'seq2seq'
    SEQ_CLS = 'seq_cls'
    ENC = 'encoder'


class LigerKernelSettings(ExtraFieldsNotAllowedBaseModel):
    use_rope: bool = True
    use_cross_entropy: bool = True
    use_geglu: bool = True


class PreTrainedModelSettings(ExtraFieldsNotAllowedBaseModel):
    model_path: Path
    model_type: ModelType

    model_kwargs: dict = {}

    transformers_settings: ModelTransformersSettings

    embeddings_initialization_strategy: dict[str, str] | None = None

    liger_kernels_settings: LigerKernelSettings | None = None


class PreTrainedAdaptersModelSettings(PreTrainedModelSettings):
    adapter_path: Path
    is_trainable: bool = False


class ModelForPeftSettings(PreTrainedModelSettings):
    peft_settings: PEFT_TYPE


class PreTrainedMultiModalModel(PreTrainedAdaptersModelSettings):
    projections_path: Path
    n_modality_embeddings: int



from turbo_alignment.modeling.multimodal.projectors.c_abstractor import CAbstractor
from turbo_alignment.modeling.multimodal.projectors.llava import (
    LlavaMultiModalProjector,
)
from turbo_alignment.modeling.multimodal.projectors.registry import (
    MultiModalProjectorRegistry,
)



import torch
from accelerate import Accelerator

from turbo_alignment.common.tf.loaders import load_model, load_tokenizer
from turbo_alignment.dataset.sampling import SamplingRMDataset
from turbo_alignment.generators.rm import RMSamplingGenerator
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import RewardSettings
from turbo_alignment.settings.datasets.base import DatasetSourceSettings
from turbo_alignment.settings.generators.outputs.chat import AnswerMessage
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.REWARD)
class RewardMetric(Metric):
    def __init__(self, settings: RewardSettings) -> None:
        super().__init__(settings=settings)
        self._settings: RewardSettings = settings
        self.tokenizer = load_tokenizer(
            self._settings.tokenizer_settings,
            self._settings.model_settings,
        )

        self.model = load_model(self._settings.model_settings, self.tokenizer)
        self.model.eval()

    def compute(self, **kwargs) -> list[MetricResults]:
        dataset: SamplingRMDataset = kwargs.get('dataset', None)
        predictions: list[list[str]] = kwargs.get('predictions', None)
        accelerator: Accelerator = kwargs.get('accelerator', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if dataset is None:
            raise ValueError('dataset should not be None')

        if predictions is None:
            raise ValueError('predictions should not be None')

        self.model = accelerator.prepare_model(self.model, device_placement=True, evaluation_mode=True)
        self.model.to(accelerator.device)

        generator = RMSamplingGenerator(
            model=self.model, tokenizer=self.tokenizer, accelerator=accelerator, micro_batch=self._settings.micro_batch
        )
        answers_per_context = len(predictions[0])

        messages = [record['messages'] for record in dataset.records for _ in range(answers_per_context)]

        answers = [
            AnswerMessage(id=str(ans_idx), content=ans)
            for ctx_answers in predictions
            for ans_idx, ans in enumerate(ctx_answers)
        ]

        rm_input_records = [
            {'id': id, 'dataset_name': '', 'messages': context, 'answers': [answer]}
            for id, (context, answer) in enumerate(zip(messages, answers))
        ]

        chat_settings = dataset.settings
        chat_settings.only_answer_loss = False

        new_dataset = SamplingRMDataset(
            source=DatasetSourceSettings(name='', records_data=rm_input_records, sample_rate=1.0),
            settings=chat_settings,
            tokenizer=self.tokenizer,
        )

        generator_outputs = generator.generate_from_dataset(new_dataset)
        reward_scores = [list(output.rewards.values()) for output in generator_outputs]

        return [
            MetricResults(
                element_wise_scores=[ElementWiseScores(label=dataset_name + '@@' + 'reward', values=reward_scores)],
                need_average=need_average,
            )
            for need_average in self._settings.need_average
        ]


def compute_metrics(eval_preds) -> dict[str, float]:
    logits, labels = eval_preds
    rewards_w, rewards_l = logits[:, 0], logits[:1]

    return {
        'pair_accuracy': labels.mean(),
        'chosen_reward': rewards_w.mean(),
        'rejected_reward': rewards_l.mean(),
        'reward_difference': rewards_w.mean() - rewards_l.mean(),
    }


def pair_comparison_accuracy_score(rewards_w: torch.Tensor, rewards_l: torch.Tensor) -> float:
    pair_scores = [1 if w > l else 0 for w, l in zip(rewards_w, rewards_l)]
    return sum(pair_scores) / len(rewards_w)



from pathlib import Path

import typer

import turbo_alignment.pipelines.sampling as sampling_pipelines
from turbo_alignment.cli.app import app
from turbo_alignment.settings.pipelines.sampling import (
    RandomSamplingSettings,
    RSOSamplingSettings,
    SamplingWithRMSettings,
)


@app.command(name='rso_sample', help='RSO sampling')
def rso_sampling_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ..., '--experiment_settings_path', exists=True, help='Path to settings config file'
    )
) -> None:
    experiment_settings = RSOSamplingSettings.parse_file(experiment_settings_path)
    sampling_pipelines.RSOSamplingStrategy().run(experiment_settings)


@app.command(name='random_sample', help='Random sampling')
def random_sampling_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ..., '--experiment_settings_path', exists=True, help='Path to settings config file'
    )
) -> None:
    experiment_settings = RandomSamplingSettings.parse_file(experiment_settings_path)
    sampling_pipelines.RandomSamplingStrategy().run(experiment_settings)


@app.command(name='rm_sample', help='Reward Modelsampling')
def rm_sampling_entrypoint(
    experiment_settings_path: Path = typer.Option(
        ..., '--experiment_settings_path', exists=True, help='Path to settings config file'
    )
) -> None:
    experiment_settings = SamplingWithRMSettings.parse_file(experiment_settings_path)
    sampling_pipelines.SamplingStrategyWithRM().run(experiment_settings)



import os
from pathlib import Path
from typing import Callable

from torch.utils.data import ConcatDataset, Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase
from transformers.data.data_collator import DataCollatorMixin

from turbo_alignment.cherry_picks.chat import ChatCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.callbacks import EvaluateFirstStepCallback
from turbo_alignment.common.tf.loaders.model import load_model
from turbo_alignment.common.tf.loaders.tokenizer import load_tokenizer
from turbo_alignment.common.tf.special_tokens_setter import SpecialTokensSetter
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.chat.chat import InferenceChatDataset
from turbo_alignment.dataset.ddpo.collators import DDPODataCollator
from turbo_alignment.dataset.ddpo.ddpo import load_ddpo_datasets as load_datasets
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets.base import DatasetStrategy
from turbo_alignment.settings.pipelines.train.ddpo import DDPOTrainExperimentSettings
from turbo_alignment.settings.s3 import ExperimentMetadata
from turbo_alignment.trainers.ddpo import DDPOTrainer, DDPOTrainingArguments

logger = get_project_logger()


class TrainDDPOStrategy(BaseTrainStrategy[DDPOTrainExperimentSettings]):
    rm_tokenizer: PreTrainedTokenizerBase = None
    rm_model: PreTrainedModel = None

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: DDPOTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> ChatCherryPickCallback:
        rm_model = kwargs.get('rm_tokenizer', None)
        assert rm_model is not None, 'RM Tokenizer should not be None'

        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[InferenceChatDataset](InferenceChatDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(
                MetricSettingsRegistry.by_name(metric.type)(**metric.parameters, rm_model=rm_model)
            )
            for metric in cherry_pick_settings.metric_settings
        ]

        return ChatCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: DDPOTrainExperimentSettings) -> DDPOTrainingArguments:
        return DDPOTrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            label_names=[],
            remove_unused_columns=False,
            beta=experiment_settings.beta,
            use_ref_model=experiment_settings.use_ref_model,
            forward_kl=experiment_settings.forward_kl,
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_data_collator(
        experiment_settings: DDPOTrainExperimentSettings, tokenizer: PreTrainedTokenizerBase, **kwargs
    ) -> Callable:
        rm_tokenizer = kwargs.get('rm_tokenizer', None)
        assert rm_tokenizer is not None, 'RM Tokenizer should not be None'
        return DDPODataCollator(
            tokenizer=tokenizer,
            rm_tokenizer=rm_tokenizer,
        )

    @staticmethod
    def _get_trainer(
        training_args: DDPOTrainingArguments,
        experiment_settings: DDPOTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: Callable,
        rm_model: PreTrainedModel = None,
    ) -> DDPOTrainer:
        model.config.use_cache = not experiment_settings.trainer_settings.gradient_checkpointing

        extra_args = {'rm': rm_model}

        return DDPOTrainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            callbacks=[],
            data_collator=data_collator,
            tokenizer=tokenizer,
            **extra_args,
        )

    def run(self, experiment_settings: DDPOTrainExperimentSettings) -> None:
        training_args = self._get_training_args(experiment_settings)

        self.tokenizer = load_tokenizer(
            experiment_settings.tokenizer_settings,
            experiment_settings.model_settings,
        )

        logger.info('Chat Tokenizer is loaded!')

        self.rm_tokenizer = load_tokenizer(
            experiment_settings.rm_tokenizer_settings,
            experiment_settings.rm_settings,
        )

        logger.info('RM Tokenizer is loaded!')

        additional_special_tokens = self._get_additional_special_tokens(experiment_settings)

        logger.info(f'Special tokens: {additional_special_tokens}')
        special_tokens_setter = SpecialTokensSetter(self.rm_tokenizer, experiment_settings.special_tokens_settings)
        special_tokens_setter.set_all()
        special_tokens_setter.set_custom_tokens(additional_special_tokens)

        logger.info('RM Special tokens added!')

        logger.info(f'Special tokens: {additional_special_tokens}')
        special_tokens_setter = SpecialTokensSetter(self.tokenizer, experiment_settings.special_tokens_settings)
        special_tokens_setter.set_all()
        special_tokens_setter.set_custom_tokens(additional_special_tokens)

        logger.info('Chat Special tokens added!')

        self.model = load_model(experiment_settings.model_settings, self.tokenizer)
        self.rm_model = load_model(experiment_settings.rm_settings, self.rm_tokenizer)

        for _, param in self.rm_model.named_parameters():
            param.requires_grad = False

        logger.info('Model is loaded!')

        train_dataset: ConcatDataset = ConcatDataset(
            datasets=load_datasets(
                experiment_settings.train_dataset_settings,
                rm_tokenizer=self.rm_tokenizer,
                chat_tokenizer=self.tokenizer,
            )
        )
        val_dataset: ConcatDataset = ConcatDataset(
            datasets=load_datasets(
                experiment_settings.val_dataset_settings,
                rm_tokenizer=self.rm_tokenizer,
                chat_tokenizer=self.tokenizer,
            )
        )

        data_collator = self._get_data_collator(
            experiment_settings=experiment_settings, tokenizer=self.tokenizer, rm_tokenizer=self.rm_tokenizer
        )

        self.trainer = self._get_trainer(
            training_args=training_args,
            experiment_settings=experiment_settings,
            model=self.model,
            tokenizer=self.tokenizer,
            rm_model=self.rm_model,
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            data_collator=data_collator,
        )

        if self.trainer.accelerator.is_main_process:
            self._dataset_and_collator_sanity_check(train_dataset, data_collator)

        self._add_trainer_callbacks(experiment_settings)
        self.trainer.add_callback(EvaluateFirstStepCallback())

        os.makedirs(self.trainer.args.output_dir, exist_ok=True)
        self._save_experiment_config(
            experiment_settings, self.trainer.model, Path(self.trainer.args.output_dir) / 'experiment.config'
        )

        experiment_metadata = ExperimentMetadata()
        self._save_experiment_metadata(
            experiment_metadata, Path(self.trainer.args.output_dir) / 'experiment_metadata.json'
        )

        self.trainer.train()

        self.trainer.save_model()

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample example:\n{dataset[0]}')

        logger.info(
            'SFT Input-w check: {input_ids}'.format(
                input_ids=collator([dataset[0], dataset[1]])['sft_inputs_w']['input_ids'][0]
            )
        )
        logger.info(
            'SFT Mask-w check: {mask}'.format(
                mask=collator([dataset[0], dataset[1]])['sft_inputs_w']['attention_mask'][0]
            )
        )
        logger.info(
            'RM Input-w check: {input_ids}'.format(
                input_ids=collator([dataset[0], dataset[1]])['rm_inputs_w']['input_ids'][0]
            )
        )
        logger.info(
            'RM Mask-w check: {mask}'.format(
                mask=collator([dataset[0], dataset[1]])['rm_inputs_w']['attention_mask'][0]
            )
        )
        logger.info(
            'SFT Input-l check: {input_ids}'.format(
                input_ids=collator([dataset[0], dataset[1]])['sft_inputs_l']['input_ids'][0]
            )
        )
        logger.info(
            'SFT Mask-l check: {mask}'.format(
                mask=collator([dataset[0], dataset[1]])['sft_inputs_l']['attention_mask'][0]
            )
        )

        logger.info(
            'RM Input-l check: {input_ids}'.format(
                input_ids=collator([dataset[0], dataset[1]])['rm_inputs_l']['input_ids'][0]
            )
        )
        logger.info(
            'RM Mask-l check: {mask}'.format(
                mask=collator([dataset[0], dataset[1]])['rm_inputs_l']['attention_mask'][0]
            )
        )



from pathlib import Path

import torch
import torchaudio
from loguru import logger
from pytorchvideo.data.clip_sampling import ConstantClipsPerVideoSampler
from torchvision import transforms

from turbo_alignment.common.data.multimodal.image.base import BaseImageReader
from turbo_alignment.common.data.multimodal.registry import AudioModalityReaderRegistry
from turbo_alignment.settings.modality import ModalityReader


@AudioModalityReaderRegistry.register(ModalityReader.IMAGEBIND)
class ImageBindAudioReader(BaseImageReader):
    def __init__(
        self,
        *args,
        num_mel_bins=128,
        target_length=204,
        sample_rate=16000,
        clip_duration=2,
        clips_per_video=3,
        mean=-4.268,
        std=9.138,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.num_mel_bins = num_mel_bins
        self.target_length = target_length
        self.sample_rate = sample_rate
        self.clip_duration = clip_duration
        self.clips_per_video = clips_per_video
        self.mean = mean
        self.std = std

        self.clip_sampler = ConstantClipsPerVideoSampler(clip_duration=clip_duration, clips_per_video=clips_per_video)

    def read(self, path: str) -> torch.Tensor:
        waveform, sr = torchaudio.load(path)
        if self.sample_rate != sr:
            waveform = torchaudio.functional.resample(waveform, orig_freq=sr, new_freq=self.sample_rate)
        all_clips_timepoints = self.get_clip_timepoints(self.clip_sampler, waveform.size(1) / self.sample_rate)
        all_clips = []
        for clip_timepoints in all_clips_timepoints:
            waveform_clip = waveform[
                :,
                int(clip_timepoints[0] * self.sample_rate) : int(clip_timepoints[1] * self.sample_rate),
            ]
            waveform_melspec = self.waveform2melspec(
                waveform_clip, self.sample_rate, self.num_mel_bins, self.target_length
            )
            all_clips.append(waveform_melspec)

        normalize = transforms.Normalize(mean=self.mean, std=self.std)
        all_clips = [normalize(ac) for ac in all_clips]

        all_clips = torch.stack(all_clips, dim=0)
        return all_clips

    @staticmethod
    def get_clip_timepoints(clip_sampler, duration):
        all_clips_timepoints = []
        is_last_clip = False
        end = 0.0
        while not is_last_clip:
            start, end, _, _, is_last_clip = clip_sampler(end, duration, annotation=None)
            all_clips_timepoints.append((start, end))
        return all_clips_timepoints

    @staticmethod
    def waveform2melspec(waveform, sample_rate, num_mel_bins, target_length):
        waveform -= waveform.mean()
        fbank = torchaudio.compliance.kaldi.fbank(
            waveform,
            htk_compat=True,
            sample_frequency=sample_rate,
            use_energy=False,
            window_type='hanning',
            num_mel_bins=num_mel_bins,
            dither=0.0,
            frame_length=25,
            frame_shift=10,
        )
        fbank = fbank.transpose(0, 1)

        n_frames = fbank.size(1)
        p = target_length - n_frames

        if abs(p) / n_frames > 0.2:
            logger.warning(
                f'Large gap between audio {n_frames=} and {target_length=}. Is the audio_target_length setting correct?'
            )

        if p > 0:
            fbank = torch.nn.functional.pad(fbank, (0, p), mode='constant', value=0)
        elif p < 0:
            fbank = fbank[:, 0:target_length]

        fbank = fbank.unsqueeze(0)
        return fbank



from abc import ABC
from functools import partial

import torch
from einops import rearrange
from pydantic import BaseModel
from timm.layers import LayerNorm, LayerNorm2d
from timm.models.regnet import RegStage

from turbo_alignment.modeling.multimodal.projectors.registry import (
    MultiModalProjectorRegistry,
)
from turbo_alignment.settings.modality import ModalityProjectorType


class HoneybeeVisualProjectorConfig(BaseModel):
    projector_type: str = 'c-abs'
    initializer_range: float = 1.0
    depth: int = 3
    mlp_depth: int = 2
    hidden_size: int = 1024
    num_eos_tokens: int = 0
    pos_emb: bool = True
    feature_layer_index: int = -1  # vision feature layer index; -1: last layer

    prenorm: bool = False  # if True, LN is applied to vision features

    encoder_hidden_size: int | None = None
    output_hidden_size: int | None = None


def build_pos_embeds(config: HoneybeeVisualProjectorConfig, num_input_tokens: int, vision_hidden_size: int):
    if config.pos_emb:
        pos_emb = torch.nn.Parameter(torch.zeros(1, num_input_tokens, vision_hidden_size))
        torch.nn.init.trunc_normal_(pos_emb, mean=0.0, std=0.02)
    else:
        pos_emb = None

    return pos_emb


def build_eos_tokens(config: HoneybeeVisualProjectorConfig, output_hidden_size: int):
    # think tokens
    num_eos_tokens = config.num_eos_tokens
    if num_eos_tokens:
        eos_tokens = torch.nn.Parameter(torch.randn(1, num_eos_tokens, output_hidden_size))
        torch.nn.init.trunc_normal_(eos_tokens, mean=0.0, std=config.initializer_range)
    else:
        eos_tokens = None

    return eos_tokens


def build_prenorm(config: HoneybeeVisualProjectorConfig):
    if getattr(config, 'prenorm', False):
        prenorm = LayerNorm(config.encoder_hidden_size)
    else:
        prenorm = None
    return prenorm


def build_mlp(depth: int, hidden_size: int, output_hidden_size: int):
    layers: list[torch.nn.Module] = [torch.nn.Linear(hidden_size, output_hidden_size)]
    for _ in range(1, depth):
        layers.append(torch.nn.SiLU())
        layers.append(torch.nn.Linear(output_hidden_size, output_hidden_size))
    return torch.nn.Sequential(*layers)


class Projector(torch.nn.Module):
    """Base projector class"""

    def __init__(
        self,
        encoder_hidden_size: int,
        text_hidden_size: int,
        n_modality_embs: int,
        config: HoneybeeVisualProjectorConfig = HoneybeeVisualProjectorConfig(),
    ):
        super().__init__()
        self.config = config

        self.config.output_hidden_size = text_hidden_size
        self.config.encoder_hidden_size = encoder_hidden_size

        # think tokens
        self.eos_tokens = build_eos_tokens(config, self.config.output_hidden_size)
        self.num_input_tokens = n_modality_embs

        # pos emb
        self.pos_emb = build_pos_embeds(config, self.num_input_tokens, self.config.encoder_hidden_size)

        self.prenorm = build_prenorm(config)

        self.build_net()

    def build_net(self):
        raise NotImplementedError()

    def _forward(self, x):
        raise NotImplementedError()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: (B, L, encoder_hidden_size) tensor from the visual backbone (CLIP visual encoder),
                including cls token.
        """
        if self.prenorm is not None:
            x = self.prenorm(x)

        if self.pos_emb is not None:
            x += self.pos_emb

        x = self._forward(x)  # (B, L, output_hidden_size)

        B = x.size(0)
        if self.eos_tokens is not None:
            x = torch.cat([x, self.eos_tokens.expand(B, -1, -1)], dim=1)

        return x

    def _load_from_state_dict(self, state_dict, *args, **kwargs):
        # update old ckpt compatible with current code
        # pos_emb = state_dict['abstractor.pos_emb']

        pos_emb = state_dict['pos_emb']
        if pos_emb.size(1) == self.pos_emb.size(1) + 1:
            # remove obsolete first pos emb (for cls token originally)
            # state_dict['abstractor.pos_emb'] = pos_emb[:, 1:]
            state_dict['pos_emb'] = pos_emb[:, 1:]

        super()._load_from_state_dict(state_dict, *args, **kwargs)


class ConvProjector(ABC, Projector):
    def _forward(self, x):
        # x: [B, L, dim]
        hw = int(x.size(1) ** 0.5)
        x = rearrange(x, 'b (h w) d -> b d h w', h=hw, w=hw)
        x = self.net(x)
        x = rearrange(x, 'b d h w -> b (h w) d')
        x = self.readout(x)

        return x


@MultiModalProjectorRegistry.register(ModalityProjectorType.C_ABSTRACTOR)
class CAbstractor(ConvProjector):
    """C-Abstractor based on RegBlock"""

    def build_net(self):
        encoder_hidden_size = self.config.encoder_hidden_size
        hidden_size = self.config.hidden_size
        output_hidden_size = self.config.output_hidden_size
        depth = self.config.depth
        mlp_depth = self.config.mlp_depth

        n_queries = self.num_input_tokens
        assert (n_queries**0.5).is_integer(), 'n_queries must be square number'
        hw = int(n_queries**0.5)

        RegBlock = partial(
            RegStage,
            stride=1,
            dilation=1,
            act_layer=torch.nn.SiLU,
            norm_layer=LayerNorm2d,
        )

        s1 = RegBlock(
            depth,
            encoder_hidden_size,
            hidden_size,
        )
        sampler = torch.nn.AdaptiveAvgPool2d((hw, hw))
        s2 = RegBlock(
            depth,
            hidden_size,
            hidden_size,
        )

        if depth:
            self.net = torch.nn.Sequential(s1, sampler, s2)
            self.readout = build_mlp(mlp_depth, hidden_size, output_hidden_size)
        else:
            self.net = sampler
            self.readout = build_mlp(mlp_depth, encoder_hidden_size, output_hidden_size)



from enum import Enum
from typing import Literal

from pydantic import Field

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel
from turbo_alignment.settings.cherry_pick import ChatCherryPickSettings
from turbo_alignment.settings.datasets.pair_preference import (
    PairPreferenceMultiDatasetSettings,
)
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings
from turbo_alignment.settings.tf.trainer import TrainerSettings


class DPOLossesType(str, Enum):
    SIGMOID = 'sigmoid'
    SIGMOID_WITH_MARGIN = 'sigmoid_with_margin'
    HINGE = 'hinge'
    IPO = 'ipo'
    KTO = 'kto'
    SLIC_HF = 'slic_hf'
    CPO = 'cpo'
    ORPO = 'orpo'
    SIMPO = 'simpo'
    APO_ZERO = 'apo_zero'
    APO_DOWN = 'apo_down'
    ASFT = 'asft'


class DPOLossSettings(ExtraFieldsNotAllowedBaseModel):
    loss_type: DPOLossesType
    beta: float = 0.1


class KTOLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.KTO]


class SigmoidLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.SIGMOID]
    label_smoothing: float = 0


class SigmoidLossWithMarginSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.SIGMOID_WITH_MARGIN]


class HingeLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.HINGE]
    norm: bool = True


class IPOLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.IPO]


class CPOLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.CPO]
    norm: bool = True


class SlicHfLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.SLIC_HF]
    delta: float = 1.0
    lam: float = 0.1
    norm: bool = False


class SimPOLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.SIMPO]
    gamma: float = 0.1


class ORPOLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.ORPO]


class ASFTLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.ASFT]


class APOZeroLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.APO_ZERO]


class APODownLossSettings(DPOLossSettings):
    loss_type: Literal[DPOLossesType.APO_DOWN]


class SyncRefModelSettings(ExtraFieldsNotAllowedBaseModel):
    sync_ref_model: bool = False
    alpha: float = 1.0
    sync_steps: int = 1


class DPOTrainerSettings(TrainerSettings):
    loss_settings: (
        SigmoidLossSettings
        | HingeLossSettings
        | IPOLossSettings
        | KTOLossSettings
        | CPOLossSettings
        | ORPOLossSettings
        | ASFTLossSettings
        | SimPOLossSettings
        | SlicHfLossSettings
        | SigmoidLossWithMarginSettings
        | APOZeroLossSettings
        | APODownLossSettings
    )
    sync_ref_settings: SyncRefModelSettings
    use_ref_model: bool = True
    use_sft_model: bool = False
    average_log_prob: bool = Field(default=False, description='Normalize log probability by length or not')


class DPOTrainExperimentSettings(BaseTrainExperimentSettings):
    train_dataset_settings: PairPreferenceMultiDatasetSettings
    val_dataset_settings: PairPreferenceMultiDatasetSettings

    cherry_pick_settings: ChatCherryPickSettings

    trainer_settings: DPOTrainerSettings



import torch
from transformers import (
    DataCollatorWithPadding,
    PreTrainedModel,
    PreTrainedTokenizerBase,
)

from turbo_alignment.dataset.chat import InferenceChatDataset
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import RetrievalUtilitySettings
from turbo_alignment.metrics.utils import calculate_cross_entropy
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


@Metric.register(MetricType.RETRIEVAL_UTILITY)
class RetrievalUtilityMetric(Metric):
    last_tokens_string: str | None = None

    def __init__(self, settings: RetrievalUtilitySettings) -> None:
        super().__init__(settings=settings)
        self._settings: RetrievalUtilitySettings = settings
        self.doc_sep_symbol = settings.doc_sep_symbol

    def compute(self, **kwargs) -> list[MetricResults]:
        model: PreTrainedModel = kwargs.get('model', None)
        tokenizer: PreTrainedTokenizerBase = kwargs.get('tokenizer', None)

        input_ids_list: list[torch.Tensor] = kwargs.get('input_token_ids', None)
        # references - ground truth question answers
        references: list[list[str]] = kwargs.get('references', None)
        dataset: InferenceChatDataset = kwargs.get('dataset', None)
        dataset_name: str = kwargs.get('dataset_name', '')

        if references is None:
            raise ValueError('references should not be None')

        if model is None:
            raise ValueError('model should not be None')

        if tokenizer is None:
            raise ValueError('tokenizer should not be None')

        if input_ids_list is None:
            raise ValueError('input_ids should not be None')

        if dataset is None:
            raise ValueError('dataset should not be None')

        retrieval_utility_scores = []

        device = model.device

        try:
            prompt_template = dataset.settings.prompt_template
            formatted_prefix = prompt_template.prefix_template.format(role=prompt_template.role_tag_mapping['bot'])
            self.last_tokens_string = f'{prompt_template.suffix_template} {formatted_prefix}'
        except (AttributeError, KeyError) as e:
            raise ValueError(
                'prompt_template should have prefix_template, suffix_template, and bot_role as spec_token'
            ) from e

        for input_ids, ref in zip(input_ids_list, references):
            true_ref = ref[0]

            input_ids = input_ids.to(device)
            segments = self._split_input_ids(tokenizer, input_ids, device)
            data_collator = DataCollatorWithPadding(tokenizer=tokenizer, return_tensors='pt')
            batch = data_collator([{'input_ids': segment} for segment in segments])
            segments_padded = batch['input_ids'].to(device)
            attention_mask_segments_padded = torch.where(segments_padded != tokenizer.pad_token_id, 1, 0).to(device)

            labels = tokenizer.encode(true_ref, return_tensors='pt').to(device)  # тут только ответ бота
            labels = labels.repeat(segments_padded.shape[0], 1)

            with torch.no_grad():
                inputs = torch.cat((segments_padded, labels), dim=1)
                attention_mask = torch.cat((attention_mask_segments_padded, torch.ones_like(labels)), dim=1)
                logits = model(inputs, attention_mask=attention_mask).logits

            labels = self._padding(labels, segments_padded, tokenizer, device)

            segments_ce = (
                calculate_cross_entropy(logits, labels, tokenizer.pad_token_id, reduction='none')
                .detach()
                .cpu()
                .tolist()
            )

            only_query_loss = segments_ce[0]
            retrieval_utility = [only_query_loss - x for x in segments_ce[1:]]
            retrieval_utility_scores.append(retrieval_utility)

        # понадобится, если будем фильтровать заретривленные доки по скору
        max_length = max(len(sublist) for sublist in retrieval_utility_scores)

        transposed_lists = [
            [sublist[i] if i < len(sublist) else 0 for sublist in retrieval_utility_scores] for i in range(max_length)
        ]

        metric_results_list = []

        for i, values in enumerate(transposed_lists, start=0):
            if i == 0:
                label = f'{dataset_name}@@docsAll_ReU'
            else:
                label = f'{dataset_name}@@docs{i}_ReU'

            metric_results_list.extend(
                [
                    MetricResults(
                        element_wise_scores=[ElementWiseScores(label=label, values=values)],
                        need_average=need_average,
                    )
                    for need_average in self._settings.need_average
                ]
            )

        return metric_results_list

    def _padding(self, labels: torch.Tensor, segment, tokenizer, device):
        padding = torch.full(
            [segment.shape[0], segment.shape[1]],
            fill_value=tokenizer.pad_token_id,
            dtype=torch.long,
        ).to(device)

        labels = torch.cat((padding, labels), dim=1)

        return labels

    def _split_input_ids(self, tokenizer, input_ids, device):
        segments = []

        last_tokens = torch.tensor(tokenizer.encode(self.last_tokens_string, add_special_tokens=False)).to(device)

        doc_sep_token_id = tokenizer.convert_tokens_to_ids(self.doc_sep_symbol)
        input_ids_flat = input_ids.squeeze()

        sep_positions = list((input_ids_flat == doc_sep_token_id).nonzero(as_tuple=True)[0].cpu())
        input_ids_question = input_ids_flat[: sep_positions[0]]

        segments.append(torch.cat((input_ids_question, last_tokens)))
        segments.append(input_ids_flat)

        # вопрос + док_i + special_tokens
        input_ids_question_docs = [
            torch.cat(
                (
                    input_ids_question,
                    input_ids_flat[sep_positions[i] : sep_positions[i + 1]],
                    last_tokens,
                )
            )
            for i in range(len(sep_positions) - 1)
        ]

        input_ids_question_docs.append(torch.cat((input_ids_question, input_ids_flat[sep_positions[-1] :])))

        segments.extend(input_ids_question_docs)
        return segments



from peft import (
    LoraConfig,
    PeftType,
    PrefixTuningConfig,
    PromptEncoderConfig,
    PromptTuningConfig,
)
from transformers import (
    AutoModel,
    AutoModelForCausalLM,
    AutoModelForSequenceClassification,
)

from turbo_alignment.common.registry import Registrable
from turbo_alignment.settings.model import ModelType


class TransformersAutoModelRegistry(Registrable):
    ...


class PeftConfigRegistry(Registrable):
    ...


TransformersAutoModelRegistry.register(ModelType.CAUSAL)(AutoModelForCausalLM)
TransformersAutoModelRegistry.register(ModelType.SEQ_CLS)(AutoModelForSequenceClassification)
TransformersAutoModelRegistry.register(ModelType.ENC)(AutoModel)

PeftConfigRegistry.register(PeftType.LORA)(LoraConfig)
PeftConfigRegistry.register(PeftType.PREFIX_TUNING)(PrefixTuningConfig)
PeftConfigRegistry.register(PeftType.PROMPT_TUNING)(PromptTuningConfig)
PeftConfigRegistry.register(PeftType.P_TUNING)(PromptEncoderConfig)



from collections import defaultdict

from nltk.tokenize import wordpunct_tokenize
from rouge_score import rouge_scorer
from rouge_score.tokenizers import Tokenizer

from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import RougeSettings
from turbo_alignment.settings.metric import ElementWiseScores, MetricResults, MetricType


class CustomTokenizer(Tokenizer):
    def tokenize(self, text: str) -> list[str]:
        return wordpunct_tokenize(text)


@Metric.register(MetricType.ROUGE)
class RougeMetric(Metric):
    def __init__(self, settings: RougeSettings) -> None:
        super().__init__(settings=settings)
        self._settings: RougeSettings = settings
        rouge_types = []

        if settings.need_compute_rouge_l:
            rouge_types.append('rougeL')

        for n in settings.need_compute_rouge_n:
            rouge_types.append(f'rouge{n}')

        self._rouge_scorer = rouge_scorer.RougeScorer(rouge_types=rouge_types, tokenizer=CustomTokenizer())

    def compute(self, **kwargs) -> list[MetricResults]:
        references: list[str] = kwargs.get('references', None)
        predictions: list[str] = kwargs.get('predictions', None)

        element_wise_scores = self._compute_element_wise(references=references, predictions=predictions)

        return [
            MetricResults(element_wise_scores=element_wise_scores, need_average=need_average)
            for need_average in self._settings.need_average
        ]

    def _compute_element_wise(self, references: list[str], predictions: list[str]) -> list[ElementWiseScores]:
        element_wise_scores = []
        metric_label_to_element_wise_scores = defaultdict(list)
        for reference_batch, prediction_batch in zip(references, predictions):
            for reference, prediction in zip(reference_batch, prediction_batch):
                scores = self._rouge_scorer.score(
                    target=reference,
                    prediction=prediction,
                )

                for label, value in scores.items():
                    metric_label_to_element_wise_scores[f'{label}_precision'].append(value.precision)
                    metric_label_to_element_wise_scores[f'{label}_recall'].append(value.recall)
                    metric_label_to_element_wise_scores[f'{label}_fmeasure'].append(value.fmeasure)

        for label, values in metric_label_to_element_wise_scores.items():
            element_wise_scores.append(ElementWiseScores(label=label, values=values))

        return element_wise_scores



from .base import InferenceExperimentSettings
from .chat import ChatInferenceExperimentSettings
from .multimodal import MultimodalInferenceExperimentSettings
from .rag import RAGInferenceExperimentSettings



from typing import Literal

from turbo_alignment.settings.datasets.base import (
    BaseDatasetSettings,
    DatasetType,
    MultiDatasetSettings,
)
from turbo_alignment.settings.datasets.chat import ChatDatasetSettings


class PairPreferenceDatasetSettings(BaseDatasetSettings):
    dataset_type: Literal[DatasetType.PAIR_PREFERENCES] = DatasetType.PAIR_PREFERENCES
    chat_settings: ChatDatasetSettings
    add_labels: bool = True


class PairPreferenceMultiDatasetSettings(PairPreferenceDatasetSettings, MultiDatasetSettings):
    ...



from typing import Callable

import torch.nn
from torch.utils.data import Dataset
from transformers import PreTrainedModel, PreTrainedTokenizerBase, TrainingArguments
from transformers.data.data_collator import DataCollatorMixin

from turbo_alignment.cherry_picks.multimodal import MultimodalCherryPickCallback
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.common.tf.loaders import load_model
from turbo_alignment.constants import TRAINER_LOGS_FOLDER
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.dataset.multimodal import InferenceMultimodalDataset
from turbo_alignment.dataset.multimodal.collators import DataCollatorWithModalityInputs
from turbo_alignment.metrics.metric import Metric
from turbo_alignment.metrics.registry import MetricSettingsRegistry
from turbo_alignment.modeling.multimodal.lm.projection import (
    ProjectionMultiModalModeling,
)
from turbo_alignment.pipelines.mixin.multimodal import MultimodalMixin
from turbo_alignment.pipelines.train.base import BaseTrainStrategy
from turbo_alignment.settings.datasets import DatasetStrategy
from turbo_alignment.settings.pipelines.train.multimodal import (
    MultimodalTrainExperimentSettings,
)
from turbo_alignment.trainers.multimodal import TrainerCustomSave

logger = get_project_logger()


class TrainMultimodalStrategy(MultimodalMixin, BaseTrainStrategy[MultimodalTrainExperimentSettings]):
    @staticmethod
    def _get_data_collator(
        experiment_settings: MultimodalTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> Callable:
        return DataCollatorWithModalityInputs(tokenizer, pad_to_multiple_of=8)

    @staticmethod
    def _get_cherry_pick_callback(
        experiment_settings: MultimodalTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
        **kwargs,
    ) -> MultimodalCherryPickCallback:
        cherry_pick_settings = experiment_settings.cherry_pick_settings

        cherry_pick_datasets = DatasetLoader[InferenceMultimodalDataset](InferenceMultimodalDataset).load_datasets(
            cherry_pick_settings.dataset_settings, tokenizer=tokenizer, strategy=DatasetStrategy.INFERENCE
        )

        metrics = [
            Metric.by_name(metric.type)(MetricSettingsRegistry.by_name(metric.type)(**metric.parameters))
            for metric in cherry_pick_settings.metric_settings
        ]

        return MultimodalCherryPickCallback(
            cherry_pick_settings=cherry_pick_settings,
            datasets=cherry_pick_datasets,
            metrics=metrics,
        )

    @staticmethod
    def _get_training_args(experiment_settings: MultimodalTrainExperimentSettings) -> TrainingArguments:
        return TrainingArguments(
            output_dir=str(experiment_settings.log_path / TRAINER_LOGS_FOLDER),
            **experiment_settings.trainer_settings.dict(),
        )

    @staticmethod
    def _get_trainer(
        training_args: TrainingArguments,
        experiment_settings: MultimodalTrainExperimentSettings,
        model: PreTrainedModel,
        tokenizer: PreTrainedTokenizerBase,
        train_dataset: Dataset,
        val_dataset: Dataset,
        data_collator: DataCollatorMixin,
        **_kwargs,
    ):
        return TrainerCustomSave(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
            callbacks=[],
            data_collator=data_collator,
            tokenizer=tokenizer,
        )

    @staticmethod
    def _load_model(
        experiment_settings: MultimodalTrainExperimentSettings,
        tokenizer: PreTrainedTokenizerBase,
    ) -> torch.nn.Module | PreTrainedModel:
        language_model = load_model(experiment_settings.model_settings, tokenizer)

        modality_encoders = TrainMultimodalStrategy._load_modality_encoders(
            experiment_settings.modality_encoder_settings_mapping,
            device=language_model.device,
            dtype=language_model.dtype,
        )

        return ProjectionMultiModalModeling(
            language_model=language_model,
            encoders=modality_encoders,
            n_modality_embs=experiment_settings.train_dataset_settings.n_modality_embeddings,
            modality_projector_mapping=experiment_settings.modality_projector_mapping,
            modality_projector_initialization_mapping=experiment_settings.modality_projector_initialization_mapping,
            peft=True,
        )

    def _dataset_and_collator_sanity_check(self, dataset: Dataset, collator: DataCollatorMixin) -> None:
        logger.info(f'Train sample example:\n{dataset[0]}')
        logger.info(
            'Input ids check: {input_ids}'.format(input_ids=collator([dataset[0], dataset[1]])['input_ids'][0])
        )
        logger.info('Mask check: {mask}'.format(mask=collator([dataset[0], dataset[1]])['attention_mask'][0]))
        logger.info('Labels check: {labels}'.format(labels=collator([dataset[0], dataset[1]])['labels'][0]))



# mypy: ignore-errors
# pylint: skip-file
import operator

import torch
import torch.nn as nn
import triton
import triton.language as tl

from turbo_alignment.modeling.liger_kernels.utils import (
    calculate_settings,
    compare_version,
    ensure_contiguous,
)

if compare_version('triton', operator.ge, '3.0.0'):
    try:
        from triton.language.extra.libdevice import tanh
    except ModuleNotFoundError:
        from triton.language.extra.cuda.libdevice import tanh
else:
    from triton.language.math import tanh


@triton.jit
def _geglu_tanh_forward_kernel(a, b, c, stride, n_cols: tl.constexpr, BLOCK_SIZE: tl.constexpr):
    program_id = tl.program_id(0)

    a += program_id * stride
    b += program_id * stride
    c += program_id * stride

    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols
    a_row = tl.load(a + col_offsets, mask=mask, other=0).to(tl.float32)
    b_row = tl.load(b + col_offsets, mask=mask, other=0)

    sqrt_2_over_pi = 0.7978845608028654
    a_cubed = a_row * a_row * a_row
    tanh_arg = sqrt_2_over_pi * (a_row + 0.044715 * a_cubed)
    tanh_result = tanh(tanh_arg)
    geglu_a = 0.5 * a_row * (1 + tanh_result)
    c_row = geglu_a * b_row
    tl.store(c + col_offsets, c_row, mask=mask)


@triton.jit
def _geglu_tanh_backward_kernel(dc, a, b, stride, n_cols: tl.constexpr, BLOCK_SIZE: tl.constexpr):
    program_id = tl.program_id(0)

    dc += program_id * stride
    a += program_id * stride
    b += program_id * stride

    col_offsets = tl.arange(0, BLOCK_SIZE)
    mask = col_offsets < n_cols

    dc_row = tl.load(dc + col_offsets, mask=mask, other=0)
    a_row = tl.load(a + col_offsets, mask=mask, other=0).to(tl.float32)
    b_row = tl.load(b + col_offsets, mask=mask, other=0)

    sqrt_2_over_pi = 0.7978845608028654
    a_cubed = a_row * a_row * a_row
    tanh_arg = sqrt_2_over_pi * (a_row + 0.044715 * a_cubed)
    tanh_result = tanh(tanh_arg)
    geglu_a = 0.5 * a_row * (1 + tanh_result)

    db_row = dc_row * geglu_a

    term1 = 0.5 * (1 + tanh_result)
    tanh_sq = tanh_result * tanh_result
    term2 = 0.5 * a_row * (1 - tanh_sq) * (sqrt_2_over_pi * (1 + 3 * 0.044715 * a_row * a_row))
    da_row = dc_row * b_row * (term1 + term2)

    tl.store(a + col_offsets, da_row, mask=mask)
    tl.store(b + col_offsets, db_row, mask=mask)


def geglu_forward(a, b):
    ori_shape = a.shape

    n_cols = ori_shape[-1]
    a = a.view(-1, n_cols)
    b = b.view(-1, n_cols)
    c = torch.empty_like(a)
    n_rows = a.shape[0]

    BLOCK_SIZE, num_warps = calculate_settings(n_cols)

    _geglu_tanh_forward_kernel[(n_rows,)](
        a,
        b,
        c,
        c.stride(-2),
        n_cols=n_cols,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=num_warps,
    )
    return a, b, c.view(*ori_shape)


def geglu_backward(a, b, dc):
    ori_shape = dc.shape
    n_cols = ori_shape[-1]
    dc = dc.view(-1, n_cols)
    n_rows = dc.shape[0]

    BLOCK_SIZE, num_warps = calculate_settings(n_cols)

    _geglu_tanh_backward_kernel[(n_rows,)](
        dc,
        a,
        b,
        dc.stride(-2),
        n_cols=n_cols,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=num_warps,
    )

    return a.view(*ori_shape), b.view(*ori_shape)


class LigerGELUMulFunction(torch.autograd.Function):
    @staticmethod
    @ensure_contiguous
    def forward(ctx, a, b):
        a, b, c = geglu_forward(a, b)
        ctx.save_for_backward(a, b)
        return c

    @staticmethod
    @ensure_contiguous
    def backward(ctx, dc):
        a, b = ctx.saved_tensors
        a, b = geglu_backward(a, b, dc)
        return a, b


class LigerGEGLUMLP(nn.Module):
    def __init__(self, config):
        super().__init__()
        self.config = config
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)

    def forward(self, x):
        return self.down_proj(LigerGELUMulFunction.apply(self.gate_proj(x), self.up_proj(x)))



from __future__ import annotations

import logging
import os
import sys

import loguru

config = {
    'handlers': [
        {
            'sink': sys.stdout,
            'serialize': False,
            'colorize': True,
            'format': '{file}:{line} [<level>{level}</level>] {time:YYYY-MM-DDTHH:mm:ssZ} {message}',
            'level': os.getenv('LOGGING_LEVEL', 'INFO'),
        }
    ]
}


def get_project_logger() -> loguru.Logger:
    from loguru import logger

    logging.getLogger('alignment-logger')

    logger.configure(**config)  # type: ignore[arg-type]

    return logger



from .classification import TrainClassificationStrategy
from .ddpo import TrainDDPOStrategy
from .dpo import TrainDPOStrategy
from .kto import TrainKTOStrategy
from .multimodal import TrainMultimodalStrategy
from .rag import TrainRAGStrategy
from .rm import TrainRMStrategy
from .sft import TrainSFTStrategy



from abc import ABC

import torch

from turbo_alignment.modeling.common.helpers import LearnableLogitScaling, Normalize
from turbo_alignment.modeling.imagebind.models import (
    ImageBindArchitectureSettings,
    ModalityType,
)


class Postprocessors(ABC, torch.nn.ModuleDict):
    def __init__(self, _settings: ImageBindArchitectureSettings):
        super().__init__()

        self[ModalityType.VISION] = Normalize(dim=-1)

        self[ModalityType.TEXT] = torch.nn.Sequential(Normalize(dim=-1), LearnableLogitScaling(learnable=True))

        self[ModalityType.AUDIO] = torch.nn.Sequential(
            Normalize(dim=-1),
            LearnableLogitScaling(logit_scale_init=20.0, learnable=False),
        )



from turbo_alignment.common.s3.checkpoints_handler import S3CheckpointsHandler
from turbo_alignment.common.tf.callbacks import CheckpointUploaderCallback
from turbo_alignment.settings.s3 import (
    CheckpointUploaderCallbackParameters,
    S3HandlerParameters,
)


class S3Mixin:
    @staticmethod
    def _get_s3_handler(s3_handler_parameters: S3HandlerParameters) -> S3CheckpointsHandler:
        return S3CheckpointsHandler(parameters=s3_handler_parameters)

    @staticmethod
    def _get_checkpoint_uploader_callback(
        s3_handler: S3CheckpointsHandler, checkpoint_uploader_callback_parameters: CheckpointUploaderCallbackParameters
    ) -> CheckpointUploaderCallback:
        checkpoint_uploader_callback = CheckpointUploaderCallback(
            parameters=checkpoint_uploader_callback_parameters, s3_handler=s3_handler
        )
        if s3_handler.directory_exists(directory_path=checkpoint_uploader_callback_parameters.directory):
            raise ValueError(f'🍅 Directory {checkpoint_uploader_callback_parameters.directory} exists in bucket')

        return checkpoint_uploader_callback



from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import pandas as pd
from clearml import Task
from transformers import (
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
)
from wandb.sdk.lib.disabled import RunDisabled
from wandb.sdk.wandb_run import Run

import wandb
from turbo_alignment.common.logging import get_project_logger

logger = get_project_logger()


class LoggingCallback(TrainerCallback, ABC):
    def on_log(
        self,
        _args: TrainingArguments,
        state: TrainerState,
        _control: TrainerControl,
        **kwargs,
    ) -> None:
        logs = kwargs.get('logs', {})
        self._log(logs=logs, state=state)

    @abstractmethod
    def _log(self, logs: dict[str, Any], state: TrainerState) -> None:
        ...

    @staticmethod
    def _rewrite_logs(logs: dict[str, Any]) -> dict[str, Any]:
        rewritten_logs = {}
        eval_prefix = 'eval_'
        eval_prefix_len = len(eval_prefix)
        test_prefix = 'test_'
        test_prefix_len = len(test_prefix)
        train_prefix = 'train_'
        train_prefix_len = len(train_prefix)
        cherry_pick_prefix = 'cherry_pick_'
        cherry_pick_prefix_len = len(cherry_pick_prefix)
        for k, v in logs.items():
            if isinstance(v, pd.DataFrame):
                v = wandb.Table(dataframe=v)

            if k.startswith(eval_prefix):
                rewritten_logs['eval/' + k[eval_prefix_len:]] = v
            elif k.startswith(test_prefix):
                rewritten_logs['test/' + k[test_prefix_len:]] = v
            elif k.startswith(train_prefix):
                rewritten_logs['train/' + k[train_prefix_len:]] = v
            elif k.startswith(cherry_pick_prefix):
                rewritten_logs['cherry_pick/' + k[cherry_pick_prefix_len:]] = v
            else:
                rewritten_logs['train/' + k] = v

        return rewritten_logs


class WandbLoggingCallback(LoggingCallback):
    def __init__(self, wandb_run: Run | RunDisabled) -> None:
        super().__init__()
        self._wandb_run = wandb_run

    def _log(self, logs: dict[str, Any], state: TrainerState) -> None:
        rewritten_logs: dict[str, Any] = self._rewrite_logs(logs)
        self._wandb_run.log({**rewritten_logs, 'train/global_step': state.global_step}, step=state.global_step)


class ClearMLLoggingCallback(LoggingCallback):
    def __init__(self, clearml_task: Task) -> None:
        super().__init__()
        self._clearml_task = clearml_task

    def _log(self, logs: dict[str, Any], state: TrainerState) -> None:
        rewritten_logs: dict[str, Any] = self._rewrite_logs(logs)

        single_value_scalars: list[str] = [
            'train_runtime',
            'train_samples_per_second',
            'train_steps_per_second',
            'train_loss',
            'epoch',
        ]

        for k, v in rewritten_logs.items():
            title, series = k.split('/')[0], '/'.join(k.split('/')[1:])

            if isinstance(v, (int, float, np.floating, np.integer)):
                if k in single_value_scalars:
                    self._clearml_task.get_logger().report_single_value(name=k, value=v)
                else:
                    self._clearml_task.get_logger().report_scalar(
                        title=title,
                        series=series,
                        value=v,
                        iteration=state.global_step,
                    )
            elif isinstance(v, pd.DataFrame):
                self._clearml_task.get_logger().report_table(
                    title=title,
                    series=series,
                    table_plot=v,
                    iteration=state.global_step,
                )
            else:
                logger.warning(
                    'Trainer is attempting to log a value of '
                    f'"{v}" of type {type(v)} for key "{k}". '
                    "This invocation of ClearML logger's function "
                    'is incorrect so this attribute was dropped. '
                )



from turbo_alignment.dataset.base.models import DatasetRecord
from turbo_alignment.dataset.chat.models import ChatMessage


class KTODatasetRecord(DatasetRecord):
    context: list[ChatMessage]
    answer: ChatMessage
    is_desirable: bool



from .chat import ChatCherryPickCallback
from .classification import ClassificationCherryPickCallback
from .multimodal import MultimodalCherryPickCallback
from .rag import RagCherryPickCallback
from .rm import RmCherryPickCallback



from turbo_alignment.settings.cherry_pick import RMCherryPickSettings
from turbo_alignment.settings.datasets.pair_preference import (
    PairPreferenceMultiDatasetSettings,
)
from turbo_alignment.settings.pipelines.train.base import BaseTrainExperimentSettings


class RMTrainExperimentSettings(BaseTrainExperimentSettings):
    train_dataset_settings: PairPreferenceMultiDatasetSettings
    val_dataset_settings: PairPreferenceMultiDatasetSettings

    cherry_pick_settings: RMCherryPickSettings



from typing import Generic, Type, TypeVar

from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base import BaseDataset
from turbo_alignment.dataset.registry import DatasetRegistry
from turbo_alignment.settings.datasets.base import DatasetStrategy, MultiDatasetSettings

logger = get_project_logger()


DatasetType = TypeVar('DatasetType', bound=BaseDataset)


class DatasetLoader(Generic[DatasetType]):
    def __init__(self, cls: Type[DatasetType] | None = None):
        self._dataset_cls = cls

    def load_datasets(
        self,
        multi_dataset_settings: MultiDatasetSettings,
        tokenizer: PreTrainedTokenizerBase,
        strategy: DatasetStrategy,
    ) -> list[DatasetType]:
        logger.info(
            f'Loading dataset {multi_dataset_settings.dataset_type} with settings:\n{multi_dataset_settings.dict()}'
        )

        datasets: list[DatasetType] = []
        for source in multi_dataset_settings.sources:
            # Determining what type of dataset is in the nested registry by field 'strategy'

            dataset = DatasetRegistry.by_name(multi_dataset_settings.dataset_type).by_name(strategy)(
                tokenizer=tokenizer,
                source=source,
                settings=multi_dataset_settings,
            )

            if self._dataset_cls is not None:
                assert isinstance(dataset, self._dataset_cls)

            datasets.append(dataset)

        return datasets



from abc import ABC
from functools import partial

import torch

from turbo_alignment.modeling.common.helpers import EinOpsRearrange
from turbo_alignment.modeling.common.transformer import (
    MultiheadAttention,
    SimpleTransformer,
)
from turbo_alignment.modeling.imagebind.models import (
    ImageBindArchitectureSettings,
    ModalityType,
)


class Trunks(ABC, torch.nn.ModuleDict):
    @staticmethod
    def __instantiate_trunk(
        embed_dim: int,
        num_blocks: int,
        num_heads: int,
        pre_transformer_ln: bool,
        add_bias_kv: bool,
        drop_path_rate: float,
    ) -> SimpleTransformer:
        return SimpleTransformer(
            embed_dim=embed_dim,
            num_blocks=num_blocks,
            ffn_dropout_rate=0.0,
            drop_path_rate=drop_path_rate,
            attn_target=partial(
                MultiheadAttention,
                embed_dim=embed_dim,
                num_heads=num_heads,
                bias=True,
                add_bias_kv=add_bias_kv,
            ),
            pre_transformer_layer=torch.nn.Sequential(
                torch.nn.LayerNorm(embed_dim, eps=1e-6) if pre_transformer_ln else torch.nn.Identity(),
                EinOpsRearrange('b l d -> l b d'),
            ),
            post_transformer_layer=EinOpsRearrange('l b d -> b l d'),
        )

    @staticmethod
    def __get_vision_trunk(vision_embed_dim: int, vision_num_blocks: int, vision_num_heads: int) -> SimpleTransformer:
        return Trunks.__instantiate_trunk(
            vision_embed_dim,
            vision_num_blocks,
            vision_num_heads,
            pre_transformer_ln=True,
            add_bias_kv=False,
            drop_path_rate=0.0,
        )

    @staticmethod
    def __get_text_trunk(
        text_embed_dim: int,
        text_num_blocks: int,
        text_num_heads: int,
    ) -> SimpleTransformer:
        return Trunks.__instantiate_trunk(
            text_embed_dim,
            text_num_blocks,
            text_num_heads,
            pre_transformer_ln=False,
            add_bias_kv=False,
            drop_path_rate=0.0,
        )

    @staticmethod
    def __get_audio_trunk(
        audio_embed_dim: int,
        audio_num_blocks: int,
        audio_num_heads: int,
        audio_drop_path_rate: float,
    ) -> SimpleTransformer:
        return Trunks.__instantiate_trunk(
            audio_embed_dim,
            audio_num_blocks,
            audio_num_heads,
            pre_transformer_ln=False,
            add_bias_kv=True,
            drop_path_rate=audio_drop_path_rate,
        )

    def __init__(self, settings: ImageBindArchitectureSettings):
        super().__init__()

        self[ModalityType.VISION] = self.__get_vision_trunk(
            settings.vision_embed_dim, settings.vision_num_blocks, settings.vision_num_heads
        )
        self[ModalityType.TEXT] = self.__get_text_trunk(
            settings.text_embed_dim,
            settings.text_num_blocks,
            settings.text_num_heads,
        )
        self[ModalityType.AUDIO] = self.__get_audio_trunk(
            settings.audio_embed_dim,
            settings.audio_num_blocks,
            settings.audio_num_heads,
            settings.audio_drop_path_rate,
        )



# mypy: ignore-errors
# pylint: skip-file
import importlib
import inspect
import logging
from collections import defaultdict
from typing import Any, Callable, ClassVar, DefaultDict, Type, TypeVar, cast

from turbo_alignment.common.registry.checks import ConfigurationError
from turbo_alignment.common.registry.from_params import FromParams

logger = logging.getLogger(__name__)

_T = TypeVar('_T')
_RegistrableT = TypeVar('_RegistrableT', bound='Registrable')

_SubclassRegistry = dict[str, tuple[type, str | None]]


class Registrable(FromParams):
    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]] = defaultdict(dict)

    default_implementation: str | None = None

    @classmethod
    def register(
        cls, name: str, constructor: str | None = None, exist_ok: bool = False
    ) -> Callable[[Type[_T]], Type[_T]]:
        registry = Registrable._registry[cls]

        def add_subclass_to_registry(subclass: Type[_T]) -> Type[_T]:
            if name in registry:
                if exist_ok:
                    message = (
                        f'{name} has already been registered as {registry[name][0].__name__}, but '
                        f'exist_ok=True, so overwriting with {cls.__name__}'
                    )
                    logger.info(message)
                else:
                    message = (
                        f'Cannot register {name} as {cls.__name__}; '
                        f'name already in use for {registry[name][0].__name__}'
                    )
                    raise ConfigurationError(message)
            registry[name] = (subclass, constructor)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def by_name(cls: Type[_RegistrableT], name):
        logger.debug(f'instantiating registered subclass {name} of {cls}')
        subclass, constructor = cls.resolve_class_name(name)
        if not constructor:
            return cast(Type[_RegistrableT], subclass)
        return cast(Callable[..., _RegistrableT], getattr(subclass, constructor))

    @classmethod
    def resolve_class_name(cls: Type[_RegistrableT], name: str) -> tuple[Type[_RegistrableT], str | None]:
        if name in Registrable._registry[cls]:
            subclass, constructor = Registrable._registry[cls][name]
            return subclass, constructor
        if '.' in name:
            # This might be a fully qualified class name, so we'll try importing its "module"
            # and finding it there.
            parts = name.split('.')
            submodule = '.'.join(parts[:-1])
            class_name = parts[-1]

            try:
                module = importlib.import_module(submodule)
            except ModuleNotFoundError:
                raise ConfigurationError(
                    f'tried to interpret {name} as a path to a class ' f'but unable to import module {submodule}'
                )

            try:
                subclass = getattr(module, class_name)
                constructor = None
                return subclass, constructor
            except AttributeError:
                raise ConfigurationError(
                    f'tried to interpret {name} as a path to a class '
                    f'but unable to find class {class_name} in {submodule}'
                )

        raise ConfigurationError(
            "If your registered class comes from custom code, you'll need to import "
            "the corresponding modules. If you're using AllenNLP from the command-line, "
            "this is done by using the '--include-package' flag, or by specifying your imports "
            "in a '.allennlp_plugins' file. "
            'Alternatively, you can specify your choices '
            """using fully-qualified paths, e.g. {"model": "my_module.models.MyModel"} """
            'in which case they will be automatically imported correctly.'
        )

    @classmethod
    def list_available(cls) -> list[str]:
        keys = list(Registrable._registry[cls].keys())
        default = cls.default_implementation

        if default is None:
            return keys
        if default not in keys:
            raise ConfigurationError(f'Default implementation {default} is not registered')
        return [default] + [k for k in keys if k != default]

    def _to_params(self) -> dict[str, Any]:
        logger.warning(
            f"'{self.__class__.__name__}' does not implement '_to_params`. Will" f" use Registrable's `_to_params`."
        )

        mro = inspect.getmro(self.__class__)[1:]

        registered_name = None
        for parent in mro:
            try:
                registered_classes = self._registry[parent]
            except KeyError:
                continue

            for name, registered_value in registered_classes.items():
                registered_class, _ = registered_value
                if registered_class == self.__class__:
                    registered_name = name
                    break

            if registered_name is not None:
                break

        if registered_name is None:
            raise KeyError(f"'{self.__class__.__name__}' is not registered")

        parameter_dict = {'type': registered_name}

        for parameter in inspect.signature(self.__class__).parameters.values():
            if parameter.default != inspect.Parameter.empty:
                logger.debug(f'Skipping parameter {parameter.name}')
                continue

            if hasattr(self, parameter.name):
                parameter_dict[parameter.name] = getattr(self, parameter.name)
            elif hasattr(self, f'_{parameter.name}'):
                parameter_dict[parameter.name] = getattr(self, f'_{parameter.name}')
            else:
                logger.warning(f'Could not find a value for positional argument {parameter.name}')
                continue

        return parameter_dict



from abc import abstractmethod
from typing import Generator, Generic, TypeVar

from accelerate import Accelerator
from accelerate.utils import set_seed
from accelerate.utils.operations import gather_object
from pydantic import BaseModel
from transformers import PreTrainedTokenizerBase

from turbo_alignment.common.data.io import write_json, write_jsonl
from turbo_alignment.common.logging import get_project_logger
from turbo_alignment.dataset.base import BaseDataset
from turbo_alignment.dataset.loader import DatasetLoader
from turbo_alignment.generators.base import BaseGenerator
from turbo_alignment.pipelines.base import BaseStrategy
from turbo_alignment.settings.datasets.base import DatasetStrategy
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput
from turbo_alignment.settings.pipelines.inference import InferenceExperimentSettings

logger = get_project_logger()

InferenceOutputRecordT = TypeVar('InferenceOutputRecordT', bound=BaseInferenceOutput)
InferenceExperimentSettingsT = TypeVar('InferenceExperimentSettingsT', bound=InferenceExperimentSettings)


class BaseInferenceStrategy(BaseStrategy, Generic[InferenceExperimentSettingsT]):
    @abstractmethod
    def _get_single_inference_settings(
        self,
        experiment_settings: InferenceExperimentSettingsT,
        accelerator: Accelerator,
    ) -> Generator[tuple[PreTrainedTokenizerBase, BaseGenerator, str, dict], None, None]:
        ...

    def run(self, experiment_settings: InferenceExperimentSettingsT) -> None:
        accelerator = Accelerator()
        set_seed(seed=0, device_specific=False)
        experiment_settings.save_path.mkdir(parents=True, exist_ok=True)

        report = {}
        for tokenizer, generator, filename, parameters_to_save in self._get_single_inference_settings(
            experiment_settings, accelerator
        ):
            datasets = DatasetLoader[BaseDataset](BaseDataset).load_datasets(
                experiment_settings.dataset_settings,
                tokenizer=tokenizer,
                strategy=DatasetStrategy.INFERENCE,
            )
            generations_output: list[BaseModel] = sum(
                [gather_object(generator.generate_from_dataset(dataset)) for dataset in datasets], []
            )

            write_jsonl(
                [out.dict(exclude_none=True) for out in generations_output], experiment_settings.save_path / filename
            )

            report[filename] = parameters_to_save

        write_json(report, experiment_settings.save_path / 'info.json')



from turbo_alignment.dataset.pair_preferences import PairPreferenceRecord
from turbo_alignment.dataset.sampling.models import SamplingDatasetRecord
from turbo_alignment.settings.generators.outputs.base import BaseInferenceOutput


class RMPairInferenceOutput(BaseInferenceOutput, PairPreferenceRecord):
    reward_w: float
    reward_l: float


class RMSamplingInferenceOutput(SamplingDatasetRecord):
    rewards: dict[str, float]



import torch
from transformers.activations import GELUActivation

from turbo_alignment.modeling.multimodal.projectors.registry import (
    MultiModalProjectorRegistry,
)
from turbo_alignment.settings.modality import ModalityProjectorType


@MultiModalProjectorRegistry.register(ModalityProjectorType.LLAVA)
class LlavaMultiModalProjector(torch.nn.Module):
    def __init__(self, encoder_hidden_size: int, text_hidden_size: int, n_modality_embs: int):
        super().__init__()
        self.encoder_hidden_size = encoder_hidden_size
        self.text_hidden_size = text_hidden_size
        self.linear_1 = torch.nn.Linear(encoder_hidden_size, text_hidden_size, bias=True)
        self.act = GELUActivation()
        self.linear_2 = torch.nn.Linear(text_hidden_size, text_hidden_size, bias=True)
        self.n_modality_embs = n_modality_embs

    def forward(self, image_features: torch.Tensor) -> torch.Tensor:
        hidden_states = self.linear_1(image_features)
        hidden_states = self.act(hidden_states)
        hidden_states = self.linear_2(hidden_states)
        return hidden_states



from enum import Enum
from pathlib import Path

from turbo_alignment.settings.base import ExtraFieldsNotAllowedBaseModel


class ModalityEncoderType(str, Enum):
    IMAGEBIND_AUDIO = 'imagebind_audio'
    IMAGEBIND_IMAGE = 'imagebind_image'
    CLIP = 'clip'
    WHISPER = 'whisper'


class ModalityEncoderSettings(ExtraFieldsNotAllowedBaseModel):
    modality_encoder_type: ModalityEncoderType
    is_pickle: bool = False
    encoder_path: Path


class Modality(str, Enum):
    IMAGE = 'image'
    AUDIO = 'audio'
    TEXT = 'text'


class ModalityReader(str, Enum):
    IMAGEBIND = 'imagebind'
    PICKLE = 'pickle'
    CLIP = 'clip'
    WHISPER = 'whisper'


class ModalityReaderSettings(ExtraFieldsNotAllowedBaseModel):
    reader_type: ModalityReader
    reader_path: Path | None


class ModalityProjectorType(str, Enum):
    LLAVA = 'llava'
    C_ABSTRACTOR = 'c_abstractor'



import torch
from peft import PeftModel, get_peft_model, prepare_model_for_int8_training
from transformers import PreTrainedModel, PreTrainedTokenizerBase
from transformers.integrations.deepspeed import is_deepspeed_zero3_enabled

from turbo_alignment.common.tf.loaders.model.registry import (
    PeftConfigRegistry,
    TransformersAutoModelRegistry,
)
from turbo_alignment.modeling.liger_kernels import apply_liger_kernel_to_gemma2
from turbo_alignment.settings.model import (
    ModelForPeftSettings,
    ModelType,
    PreTrainedAdaptersModelSettings,
    PreTrainedModelSettings,
)
from turbo_alignment.settings.tf.peft import PEFT_TYPE


def _prepare_model_for_peft(model: PreTrainedModel, peft_settings: PEFT_TYPE) -> PeftModel:
    peft_params = peft_settings.dict()
    peft_params.pop('name')

    peft_config = PeftConfigRegistry.by_name(peft_settings.name)(**peft_params)

    return get_peft_model(model, peft_config)


def _load_pretrained_adapters(
    model: PreTrainedModel,
    model_settings: PreTrainedAdaptersModelSettings,
) -> PeftModel:
    return PeftModel.from_pretrained(
        model,
        model_settings.adapter_path,
        is_trainable=model_settings.is_trainable,
    )


def unfreeze_params(layer):
    for param in layer.parameters():
        param.requires_grad = True


def load_model(
    model_settings: PreTrainedModelSettings,
    tokenizer: PreTrainedTokenizerBase,
) -> PreTrainedModel:
    if model_settings.liger_kernels_settings is not None:
        apply_liger_kernel_to_gemma2(
            rope=model_settings.liger_kernels_settings.use_rope,
            cross_entropy=model_settings.liger_kernels_settings.use_cross_entropy,
            geglu=model_settings.liger_kernels_settings.use_geglu,
        )

    model = TransformersAutoModelRegistry.by_name(model_settings.model_type).from_pretrained(
        model_settings.model_path,
        **model_settings.transformers_settings.dict(exclude_none=True),
        **model_settings.model_kwargs,
        torch_dtype=torch.bfloat16,
    )

    if model_settings.transformers_settings.load_in_8bit:
        model = prepare_model_for_int8_training(model)

    model.resize_token_embeddings(len(tokenizer))

    if model_settings.embeddings_initialization_strategy is not None:
        with torch.no_grad():
            for new_token, old_token in model_settings.embeddings_initialization_strategy.items():
                new_token_id = tokenizer.get_added_vocab()[new_token]
                old_token_id = tokenizer.encode(old_token, add_special_tokens=False)[0]

                if model.config.model_type == 'gpt_neox':
                    model.gpt_neox.embed_in.weight[new_token_id, :] = torch.clone(
                        model.gpt_neox.embed_in.weight[old_token_id, :]
                    )
                    if model_settings.model_type == 'causal':
                        model.embed_out.weight[new_token_id, :] = torch.clone(model.embed_out.weight[old_token_id, :])

                elif model.config.model_type == 'llama':
                    model.model.embed_tokens.weight[new_token_id, :] = model.model.embed_tokens.weight[old_token_id, :]

    if isinstance(model_settings, PreTrainedAdaptersModelSettings):
        model = _load_pretrained_adapters(model, model_settings)
    elif isinstance(model_settings, ModelForPeftSettings):
        # creating learnable adapters and freezing non-training parameters
        model = _prepare_model_for_peft(model, model_settings.peft_settings)

        # deepspeed stage3 is currently doens't work with seq_cls head and peft
        if model_settings.model_type == ModelType.SEQ_CLS and is_deepspeed_zero3_enabled():
            model.base_model.model.score = torch.nn.Linear(
                in_features=model.base_model.model.score.original_module.in_features,
                out_features=model.base_model.model.score.original_module.out_features,
                bias=model.base_model.model.score.original_module.bias,
            )
            model.base_model.model.score.weight.requires_grad = True

    return model
